"""
Utilities functions for GDR DAS format.

See: https://gdr.openei.org/das_data_standard for more info.
"""

import numpy as np

import dascore as dc
from dascore.core import get_coord
from dascore.utils.hdf5 import extract_h5_attrs, h5_matches_structure
from dascore.utils.misc import unbyte

# This defines the metadata/rawdata version combinations that define GDR versions.
_COMPOSITE_VERSIONS = {
    ("DAS-RCN v1.10", "PRODML v2.2"): "1",
}

_BASE_STRUCTURE = (
    "DasMetadata/Interrogator/Acquisition",
    "DasRawData/DasTimeArray",
    "DasRawData/RawData",
    "DasMetadata.MetadataStandard",
    "DasMetadata.RawDataStandard",
)

# Attribute map for version 1. {current_name: new_name}
ACQ = "DasMetadata/Interrogator/Acquisition"
_V1_ATTR_MAP = {
    f"{ACQ}.GaugeLength": "gauge_length",
    f"{ACQ}.GaugeLengthUnit": "gauge_length_units",
    f"{ACQ}.UnitOfMeasure": "data_units",
    "DasMetadata/Interrogator.SerialNumber": "instrument_id",
}


def _get_version(h5fi):
    """Get the version code of the GDR file."""
    if not h5_matches_structure(h5fi, _BASE_STRUCTURE):
        return False
    meta = h5fi["DasMetadata"].attrs
    data_fmt = meta["RawDataStandard"]
    meta_fmt = meta["MetadataStandard"]
    return "GDR_DAS", _COMPOSITE_VERSIONS[(meta_fmt, data_fmt)]


def _get_attrs_coords_and_data(resource, snap):
    """
    Get attributes, coordinates, and data from the file.
    """
    fill = {"NaN": "", "nan": ""}
    attrs = extract_h5_attrs(resource, _V1_ATTR_MAP, fill_values=fill)
    coords = _get_coord_manager(resource, snap)
    data = resource["DasRawData/RawData"]
    return attrs, coords, data


def _get_coord_manager(resource, snap=True):
    """Get a coordinate manager from the file."""

    def get_time_coord(resource, snap):
        """Get the time coordinate."""
        # TODO: I am not sure if time will always be in ns, check on it.
        time = resource["DasRawData/DasTimeArray"]
        if not snap:
            return get_coord(data=np.array(time).astype("datetime64[ns]"))
        t1 = np.int64(time[0]).astype("datetime64[ns]")
        t2 = np.int64(time[-1]).astype("datetime64[ns]")
        step = (t2 - t1) / (len(time) - 1)
        return get_coord(start=t1, stop=t2, step=step).change_length(len(time))

    def get_dist_coord(resource, length):
        """Get distance coordinates."""
        # Note: There is not enough info to correctly infer the start of
        # distance coordinate since Channels are often not included. In this
        # case we just assume the distance starts at 0 since the location of
        # each channel must be attached alter anyway. This at least includes
        # correct dx information.
        group = resource["DasMetadata/Interrogator/Acquisition"]
        dx = float(unbyte(group.attrs["SpatialSamplingInterval"]))
        units = unbyte(group.attrs["SpatialSamplingIntervalUnit"])
        start = 0
        stop = length * dx
        coord = get_coord(start=start, stop=stop, step=dx, units=units)
        return coord.change_length(length)

    def get_dims(dataset):
        """Get the dimension names."""
        das_dims = dataset.attrs["DasDimensions"]
        out = [""] * 2
        for num, dim in enumerate(das_dims):
            if dim.startswith("time"):
                out[num] = "time"
            elif dim == "locus":
                out[num] = "distance"
        assert all(out)
        return tuple(out)

    time_coord = get_time_coord(resource, snap)
    dataset = resource["DasRawData/RawData"]
    dims = get_dims(dataset)
    # Get distance coord.
    dist_axis = dims.index("distance")
    dist_length = dataset.shape[dist_axis]
    dist_coord = get_dist_coord(resource, dist_length)

    coords = {
        "time": time_coord,
        "distance": dist_coord,
    }

    return dc.get_coord_manager(coords=coords, dims=dims)


def _maybe_trim_data(cm, data, time=None, distance=None, **kwargs):
    """Maybe trim the data."""
    if time is not None or distance is not None:
        cm, data = cm.select(time=time, distance=distance, array=data)
    return cm, data



"""Tests for converting velocity to strain-rate."""

from __future__ import annotations

import itertools

import numpy as np
import pytest

import dascore as dc
from dascore.exceptions import ParameterError, PatchAttributeError
from dascore.units import get_quantity


@pytest.fixture(scope="class")
def linear_velocity_patch(random_patch):
    """A patch increasing linearly along distance dimension."""
    ind = random_patch.dims.index("distance")
    dist_len = random_patch.shape[ind]
    new = np.arange(dist_len, dtype=np.float64)
    new_data: np.ndarray = np.broadcast_to(new[:, None], random_patch.shape)
    new_attrs = {"data_type": "velocity", "data_units": "m/s"}
    # make distance and odd, non 1, value to test normalization
    dist = random_patch.get_array("distance")
    out = random_patch.new(data=new_data, attrs=new_attrs)
    return out.update_coords(distance=dist * 1.23)


class TestStrainRateConversion:
    """Tests for converting velocity to strain-rate."""

    @pytest.fixture(scope="class")
    def patch_strain_rate_default(self, terra15_das_patch):
        """Return the default terra15 converted to strain rate."""
        return terra15_das_patch.velocity_to_strain_rate()

    def test_attrs(self, patch_strain_rate_default):
        """Ensure the attributes were updated with strain_rate."""
        attrs = patch_strain_rate_default.attrs
        assert attrs["data_type"] == "strain_rate"

    def test_data_different(self, patch_strain_rate_default, terra15_das_patch):
        """The data should have changed."""
        data1 = patch_strain_rate_default.data
        data2 = terra15_das_patch.data
        assert not np.all(np.equal(data1, data2))

    def test_raises_on_strain_rate(self, patch_strain_rate_default):
        """It does not make sense to apply this twice."""
        with pytest.raises(PatchAttributeError, match="velocity"):
            _ = patch_strain_rate_default.velocity_to_strain_rate()

    def test_update_units(self, patch_strain_rate_default, terra15_das_patch):
        """Ensure units are updated. See issue #144."""
        new_units = get_quantity(patch_strain_rate_default.attrs.data_units)
        old_units = get_quantity(terra15_das_patch.attrs.data_units)
        assert new_units != old_units
        assert new_units == old_units / get_quantity("m")

    def test_odd_step_multiple_raise(self, terra15_das_patch):
        """Ensure odd step multiples raise a parameter error."""
        msg = "must be even"
        with pytest.raises(ParameterError, match=msg):
            terra15_das_patch.velocity_to_strain_rate(step_multiple=1)

    def test_step_multiple(self, terra15_das_patch):
        """Ensure strain rate multiples > 2 are supported."""
        # Note: this is a bit of a weak test. However, since we just call
        # Patch.differentiate under the hood, and that is better tested
        # for supporting different step sizes, its probably sufficient.
        out1 = terra15_das_patch.velocity_to_strain_rate(step_multiple=2)
        assert isinstance(out1, dc.Patch)
        out2 = terra15_das_patch.velocity_to_strain_rate(step_multiple=4)
        assert isinstance(out2, dc.Patch)

    def test_gauge_multiple_deprecated(self, terra15_das_patch):
        """Ensure using gauge_multiple issues deprecation warning."""
        with pytest.warns(DeprecationWarning):
            out1 = terra15_das_patch.velocity_to_strain_rate(gauge_multiple=1)
        out2 = terra15_das_patch.velocity_to_strain_rate(step_multiple=2)
        assert out1.equals(out2)

    def test_no_data_units(self, terra15_das_patch):
        """Ensure a patch with no data units still works."""
        patch = terra15_das_patch.update_attrs(data_units="").velocity_to_strain_rate()
        assert not patch.attrs.data_units

    def test_linear_patch(self, linear_velocity_patch):
        """Test conversion of analytical function."""
        step_mults = [2, 4, 6]
        order = [2, 4, 6]
        for step_mult, order in itertools.product(step_mults, order):
            if order > 2:  # Need findiff for orders > 2
                pytest.importorskip("findiff")
            step = linear_velocity_patch.get_coord("distance").step
            expected = 1.0 / step
            out1 = linear_velocity_patch.velocity_to_strain_rate(
                step_multiple=step_mult, order=order
            )
            assert np.allclose(out1.data, expected)


class TestStaggeredStrainRateConversion:
    """Tests for converting velocity to strain-rate with staggered function."""

    @pytest.fixture(scope="class")
    def patch_strain_rate_default(self, terra15_das_patch):
        """Return the default terra15 converted to strain rate."""
        return terra15_das_patch.velocity_to_strain_rate_edgeless()

    def test_attrs(self, patch_strain_rate_default):
        """Ensure the attributes were updated with strain_rate."""
        attrs = patch_strain_rate_default.attrs
        assert attrs["data_type"] == "strain_rate"

    def test_coords_odd_step(self, terra15_das_patch):
        """Ensure coords are staggered when step multiple is odd."""
        pre_dist = terra15_das_patch.get_array("distance")
        out = terra15_das_patch.velocity_to_strain_rate_edgeless(step_multiple=1)
        post_dist = out.get_array("distance")
        assert len(pre_dist) == len(post_dist) + 1
        expected = (pre_dist[1:] + pre_dist[:-1]) / 2
        assert np.allclose(post_dist, expected)

    def test_coords_even_step(self, terra15_das_patch):
        """The coords should be a subset of original when step is even."""
        pre_dist = terra15_das_patch.get_array("distance")
        out = terra15_das_patch.velocity_to_strain_rate_edgeless(step_multiple=2)
        post_dist = out.get_array("distance")
        assert len(pre_dist) == len(post_dist) + 2
        expected = pre_dist[1:-1]
        assert np.allclose(post_dist, expected)

    def test_raises_on_strain_rate(self, patch_strain_rate_default):
        """It does not make sense to apply this twice."""
        with pytest.raises(PatchAttributeError, match="velocity"):
            _ = patch_strain_rate_default.velocity_to_strain_rate_edgeless()

    def test_update_units(self, patch_strain_rate_default, terra15_das_patch):
        """Ensure units are updated. See issue #144."""
        new_units = get_quantity(patch_strain_rate_default.attrs.data_units)
        old_units = get_quantity(terra15_das_patch.attrs.data_units)
        assert new_units != old_units
        assert new_units == old_units / get_quantity("m")

    def test_no_data_units(self, terra15_das_patch):
        """Ensure a patch with no data units still works."""
        patch = terra15_das_patch.update_attrs(
            data_units=""
        ).velocity_to_strain_rate_edgeless()
        assert not patch.attrs.data_units

    def test_linear_patch(self, linear_velocity_patch):
        """Test conversion of analytical function."""
        step = linear_velocity_patch.get_coord("distance").step
        expected = 1.0 / step
        out1 = linear_velocity_patch.velocity_to_strain_rate_edgeless()
        assert np.allclose(out1.data, expected)

    def test_functions_equal(self, terra15_das_patch):
        """Ensure the functions are equal for even steps."""
        patch = terra15_das_patch
        for mult in [2, 4, 6, 8]:
            # Get function 1 output and trim off edges.
            strain1 = patch.velocity_to_strain_rate(step_multiple=mult).select(
                distance=(mult // 2, -mult // 2), samples=True
            )

            # Function 2's output should match function 1.
            strain2 = patch.velocity_to_strain_rate_edgeless(step_multiple=mult)

            assert np.allclose(strain1.data, strain2.data)



"""Tests for applying rolling functions on patch's data."""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

import dascore as dc
from dascore.exceptions import ParameterError
from dascore.units import m
from dascore.utils.misc import all_close
from dascore.utils.pd import rolling_df


@pytest.fixture(scope="class")
def range_patch(random_patch):
    """Return a patch with sequential values for its 0th dim."""
    new_data = np.ones_like(random_patch.data)
    range_array = np.arange(0, new_data.shape[0])[:, None]
    return random_patch.new(data=new_data * range_array)


class TestRolling:
    """Tests for applying rolling functions on patch's data."""

    window = 16
    step = 8

    @pytest.fixture(scope="class")
    def dist_roll_range_patch(self, range_patch):
        """Return a patch with sequential values for its 0th dim."""
        dist_step = range_patch.get_coord("distance").step
        out = range_patch.rolling(distance=dist_step * self.window).max()
        return out

    def test_apply_correct_no_step(self, dist_roll_range_patch):
        """Ensure the apply is correct without using a step size."""
        data_no_nan = dist_roll_range_patch.dropna("distance").data
        # determine what the values should be:
        ax_0_len = dist_roll_range_patch.shape[0]
        samps = self.window - 1
        expected_values = np.arange(samps, ax_0_len)[:, None]
        # and assert they are indeed that.
        assert np.allclose(data_no_nan, expected_values)

    def test_along_time(self, range_patch):
        """Ensure time axis also works."""
        time_step = range_patch.get_coord("time").step
        out = range_patch.rolling(time=time_step * self.window).sum()
        shape = range_patch.shape
        # Note: we start at 0 here because range is along 0th axis and
        # time is first axis so all windows are the same values.
        expected = (np.arange(0, shape[0]) * self.window)[:, None]
        assert np.allclose(out.dropna("time").data, expected)
        # this should also work if patch is transposed.
        trans = range_patch.transpose()
        out = trans.rolling(time=time_step * self.window).sum()
        assert np.allclose(out.dropna("time").data, expected.transpose())

    def test_rolling_timdelta(self, random_patch):
        """Ensure rolling works with timedeltas."""
        time_step = random_patch.get_coord("time").step
        time = time_step * self.window
        out1 = random_patch.rolling(time=dc.to_timedelta64(time)).sum()
        out2 = random_patch.rolling(time=time * dc.get_quantity("s")).sum()
        assert all_close(out1, out2)

    def test_apply_with_step(self, range_patch):
        """Ensure apply works with various step sizes."""
        # first calculate rolling max on time axis.
        step = range_patch.get_coord("distance").step
        roll = range_patch.rolling(distance=self.window * step, step=self.step * step)

        out = roll.max().dropna("distance")
        start = roll.get_start_index()
        # Determine what output should be.
        vals = np.arange(self.window - 1, range_patch.shape[0])
        expected = vals[start :: self.step, None]
        assert np.allclose(out.data, expected)

    def test_window_size_one(self, random_patch):
        """Ensure we can get a window size of one."""
        time_step = random_patch.get_coord("time").step
        out = random_patch.rolling(time=time_step, step=None).mean()
        # window size of 1 with no step results in the same as input.
        assert out.equals(random_patch)

    def test_1d_patch(self):
        """Ensure rolling works with 1D patch."""
        patch = dc.Patch(
            data=np.arange(10),
            coords={"time": dc.to_datetime64(np.arange(10))},
            dims=("time",),
        )
        expected = patch.data[:-2]
        roll = patch.rolling(time=3)
        out = roll.min().dropna("time")
        assert np.allclose(expected, out.data)

    def test_3d_patch(self, range_patch_3d):
        """Ensure rolling works with 3D patch."""
        patch = range_patch_3d
        # first try along time axis.
        out = patch.rolling(time=3).min().dropna("time")
        expected = np.arange(range_patch_3d.shape[0] - 2)
        assert np.allclose(out.data, expected[:, None, None])
        # then distance
        out = patch.rolling(distance=3).min().dropna("distance")
        expected = np.arange(range_patch_3d.shape[0])
        assert np.allclose(out.data, expected[:, None, None])
        # then smell
        out = patch.rolling(smell=3).min().dropna("smell")
        expected = np.arange(range_patch_3d.shape[0])
        assert np.allclose(out.data, expected[:, None, None])

    def test_window_zero_raises(self, random_patch):
        """When the window or step is entered 0 it should raise."""
        random_patch.get_coord("time")
        msg = "Window or step size can't be zero"
        with pytest.raises(ParameterError, match=msg):
            random_patch.rolling(time=0)

    def test_window_too_big_raises(self, random_patch):
        """When the window or step is too large it should raise."""
        coord = random_patch.get_coord("time")
        duration = coord.max() - coord.min()
        msg = "results in a window larger than coordinate"
        with pytest.raises(ParameterError, match=msg):
            random_patch.rolling(time=duration * 2)

    def test_center(self, random_patch):
        """Ensure the center option places NaN at start and end."""
        out = random_patch.rolling(time=1, center=True).mean()
        time_ax = out.dims.index("time")
        first_label = np.take(out.data, -1, axis=time_ax)
        assert np.all(np.isnan(first_label))
        last_label = np.take(out.data, 0, axis=time_ax)
        assert np.all(np.isnan(last_label))

    @pytest.mark.parametrize("_", list(range(5)))
    def test_compare_to_pandas(self, range_patch, _):
        """Test the apply method of PatchRoller when distance coordinate is entered
        and the first axis is distance.
        """
        random = np.random.RandomState(42)
        patch = range_patch
        axis = patch.dims.index("distance")
        # random window and step sizes for each trial run.
        window = random.randint(1, patch.shape[axis])
        step = random.randint(1, patch.shape[axis])
        # setup patch and apply mean.
        channel_spacing = patch.attrs["distance_step"]
        roller = patch.rolling(
            distance=(window * channel_spacing) * m,
            step=(step * channel_spacing) * m,
        )
        applied_result = roller.apply(np.mean).dropna("distance").data
        # do the same with pandas and compare result
        df = pd.DataFrame(patch.data)
        rolling_mean_pandas = rolling_df(df, window, step=step, axis=axis).mean()
        filtered_data_pandas = rolling_mean_pandas.dropna(axis=axis).values
        assert applied_result.shape == filtered_data_pandas.shape
        assert np.allclose(applied_result, filtered_data_pandas)

    def test_pandas_engine_raises_3d_patch(self, range_patch_3d):
        """Pandas engine should raise when required with a 4d patch."""
        msg = "Cannot use Pandas engine on patches with more than"
        with pytest.raises(ParameterError, match=msg):
            range_patch_3d.rolling(time=1, engine="pandas").mean()

    def test_misc(self, random_patch):
        """Test miscellaneous functionality."""
        time_step = random_patch.get_coord("time").step
        window = 10 * time_step
        step = 10 * time_step
        rollers = (
            random_patch.rolling(time=window, step=step, engine="numpy"),
            random_patch.rolling(time=window, step=step, engine="pandas"),
        )
        for roller in rollers:
            assert isinstance(roller.mean(), dc.Patch)
            assert isinstance(roller.median(), dc.Patch)
            assert isinstance(roller.min(), dc.Patch)
            assert isinstance(roller.max(), dc.Patch)
            assert isinstance(roller.std(), dc.Patch)
            assert isinstance(roller.sum(), dc.Patch)

    def test_pandas_apply(self, random_patch):
        """Test pandas apply works."""
        # This can be very slow so we use a large window and step size.
        dt = random_patch.get_coord("time").step
        time_len = random_patch.shape[random_patch.dims.index("time")]
        window = (time_len - 1) * dt
        step = (time_len - 1) * dt
        roll1 = random_patch.rolling(time=window, step=step, engine="pandas")
        roll2 = random_patch.rolling(time=window, step=step, engine="numpy")
        patch_1 = roll1.apply(np.sum)
        patch_2 = roll2.apply(np.sum)
        assert patch_2.shape == patch_1.shape


class TestNumpyVsPandasRolling:
    """Ensure numpy rolling return the same results as pandas rolling."""

    # window size/step in terms of steps
    combinations = (
        (1, 1),
        (120, 60),
        (100, 100),
        (13, 7),
        (40, 80),
    )

    @pytest.mark.parametrize("data", combinations)
    def test_time_dim(self, data: tuple[int, int], range_patch):
        """Ensure pandas and numpy engine return same results along time."""
        coord = range_patch.get_coord("time")
        dt = coord.step
        win, step = dt * data[0], dt * data[1]
        numpy_roll = range_patch.rolling(time=win, step=step, engine="numpy")
        pandas_roll = range_patch.rolling(time=win, step=step, engine="pandas")

        nump_out = numpy_roll.mean()
        pand_out = pandas_roll.mean()

        is_close = np.isclose(nump_out.data, pand_out.data)
        is_nan = np.isnan(nump_out.data) & np.isnan(pand_out.data)
        assert np.all(is_close | is_nan)

    @pytest.mark.parametrize("data", combinations)
    def test_dist_dim(self, data: tuple[int, int], range_patch):
        """Ensure pandas and numpy engine return same results along distance."""
        coord = range_patch.get_coord("distance")
        dt = coord.step
        win, step = dt * data[0], dt * data[1]
        numpy_roll = range_patch.rolling(distance=win, step=step, engine="numpy")
        pandas_roll = range_patch.rolling(distance=win, step=step, engine="pandas")

        nump_out = numpy_roll.mean()
        pand_out = pandas_roll.mean()

        is_close = np.isclose(nump_out.data, pand_out.data)
        is_nan = np.isnan(nump_out.data) & np.isnan(pand_out.data)
        assert np.all(is_close | is_nan)

    def test_center_same(self, range_patch):
        """Ensure center values are handled the same."""
        dt = range_patch.get_coord("time").step
        numpy_out = range_patch.rolling(time=13 * dt, center=True, engine="numpy").sum()
        pandas_out = range_patch.rolling(
            time=13 * dt, center=True, engine="pandas"
        ).sum()
        numpy_isnan = np.isnan(numpy_out.data)
        pandas_isnan = np.isnan(pandas_out.data)
        assert np.all(
            np.equal(numpy_isnan, pandas_isnan)
        ), "The NaN indices do not match"

    def test_center_same_stepped(self, range_patch):
        """Ensure center values are handled the same."""
        dt = range_patch.get_coord("time").step
        numpy_out = range_patch.rolling(
            time=13 * dt, step=3 * dt, center=True, engine="numpy"
        ).sum()
        pandas_out = range_patch.rolling(
            time=13 * dt, step=3 * dt, center=True, engine="pandas"
        ).sum()
        numpy_isnan = np.isnan(numpy_out.data)
        pandas_isnan = np.isnan(pandas_out.data)
        assert np.all(
            np.equal(numpy_isnan, pandas_isnan)
        ), "The NaN indices do not match"

    def test_dimension_order(self, range_patch):
        """Ensure the dimension order doesn't matter."""
        for patch in [range_patch, range_patch.transpose()]:
            for dim in patch.dims:
                coord = patch.get_coord(dim)
                step = coord.step
                total_len = len(coord) - 2
                kwargs_pandas = {
                    dim: step * total_len,
                    "step": total_len * step,
                    "engine": "pandas",
                }
                kwargs_numpy = {
                    dim: step * total_len,
                    "step": total_len * step,
                    "engine": "numpy",
                }
                pandas_out = patch.rolling(**kwargs_pandas).mean().dropna(dim)
                numpy_out = patch.rolling(**kwargs_numpy).mean().dropna(dim)
                assert pandas_out == numpy_out



"""
Utility functions for AP sensing module.
"""

import pandas as pd

import dascore as dc
from dascore.core import get_coord, get_coord_manager
from dascore.utils.misc import maybe_get_items

_ATTR_MAP = {
    "GaugeLength": "gauge_length",
    "SamplingFrequency[Hz]": "sampling_frequency",
    "Fibre Length Multiplier": "fiber_length_multiplier",
    "Start Distance (m)": "start_distance",
    "Stop Distance (m)": "stop_distance",
    "Fibre Length per Metre": "fiber_length_per_metre",
    "GPSTimeStamp": "gps_timestamp",
    "CPUTimeStamp": "cpu_timestamp",
    "Tags": "tag",
    "PulseWidth[ns]": "pulse_width",
    "MeasureLength[m]": "measured_length",
    "StartPosition[m]": "start_position",
    "SpatialResolution[m]": "spatial_resolution",
    # oops, they spelled information "infomation"
    "SystemInfomation.Devices1.SerialNum": "instrument_id",
}

_EXPECTED_ATTRS = set(_ATTR_MAP)


def _get_version_string(resource, version):
    """Return version string if silixa h5 format else False."""
    dataset = resource.get("Acoustic", {})
    attrs_names = set(getattr(dataset, "attrs", dataset))
    has_attrs = _EXPECTED_ATTRS.issubset(attrs_names)
    if dataset is None or not has_attrs:
        return False
    return version


def _read_time_string(time_str):
    """Read the timestring like dd/mm/yyyy."""
    out = pd.to_datetime(time_str.replace(" (UTC)", ""), dayfirst=True)
    return dc.to_datetime64(out)


def _get_time_coord(attr_dict, shape):
    """Create the time coordinate."""
    gps_time = _read_time_string(attr_dict["gps_timestamp"])
    cpu_time = _read_time_string(attr_dict["cpu_timestamp"])
    time_min = cpu_time if pd.isnull(gps_time) else gps_time
    sampling_rate = 1.0 / float(attr_dict["sampling_frequency"])
    step = dc.to_timedelta64(sampling_rate)
    length = shape[0]
    coord = get_coord(start=time_min, step=step, shape=(length,))
    return coord


def _get_distance_coord(attr_dict, data_shape):
    """Get distance coordinate from AP sensing data."""
    # Note: To be consistent with TDMS reader (see tdms.utils.get_distance_coord)
    # We use this method for calculating distance, although start distance
    # and stop distance are included.
    multiplier = float(attr_dict["fiber_length_multiplier"])
    total_length = float(attr_dict["measured_length"]) * multiplier
    start = float(attr_dict["start_position"]) + total_length
    step = float(attr_dict["spatial_resolution"]) * multiplier
    stop = start + data_shape[1] * step
    coord = get_coord(start=start, stop=stop, step=step, units="m")
    return coord.change_length(data_shape[1])


def _get_coords(attrs_dict, shape):
    """Get coordinates of AP_sensing file."""
    # first get time
    time_coord = _get_time_coord(attrs_dict, shape)
    distance_coord = _get_distance_coord(attrs_dict, shape)
    cm = get_coord_manager(
        {"time": time_coord, "distance": distance_coord},
        dims=("time", "distance"),
    )
    return cm


def _get_attr_dict(resource):
    """Get the attribute map."""
    ds = resource["Acoustic"]
    attrs_dict = maybe_get_items(ds.attrs, _ATTR_MAP)
    attrs_dict["coords"] = _get_coords(attrs_dict, ds.shape)
    return attrs_dict


def _get_attr(resource, attr_cls, extras=None):
    """Get the attribute class"""
    attrs = _get_attr_dict(resource)
    expected_fields = set(attr_cls.model_fields)
    attrs_sub = {i: v for i, v in attrs.items() if i in expected_fields}
    attrs_sub.update(extras if extras else {})
    attrs = attr_cls.model_validate(attrs_sub)
    return attrs


def _get_patch(resource, time=None, distance=None, attr_cls=dc.PatchAttrs):
    """Get a patch from ap_sensing file."""
    attrs = _get_attr_dict(resource)
    coords = attrs["coords"]
    data = resource["Acoustic"]
    if time is not None or distance is not None:
        coords, data = coords.select(array=data, time=time, distance=distance)
        attrs["coords"] = coords
    expected_fields = set(attr_cls.model_fields)
    attrs_sub = {i: v for i, v in attrs.items() if i in expected_fields}
    attrs = attr_cls.model_validate(attrs_sub)
    return dc.Patch(data=data[:], coords=coords, attrs=attrs)



"""Tests for dascore's downloader."""

from __future__ import annotations

import pandas as pd
import pytest

from dascore.utils.downloader import fetch, get_registry_df


@pytest.fixture()
def registry_df():
    """Load the registry df."""
    df = get_registry_df()
    return df


class TestRegistryDF:
    """Tests for getting the data registry."""

    def test_dataframe(self, registry_df):
        """Ensure a non-empty df was returned."""
        assert len(registry_df)
        assert isinstance(registry_df, pd.DataFrame)


class TestFetch:
    """Tests for fetching filepaths of test files."""

    def test_multiple_fetch(self, registry_df):
        """Ensure multiple fetch calls return same path."""
        path = registry_df["name"].iloc[0]
        assert fetch(path) == fetch(path)

    def test_existing_file(self, registry_df):
        """Ensure an existing file just returns."""
        path = fetch(registry_df["name"].iloc[0])
        assert fetch(path) == path



"""Core module for reading and writing DASDAE format."""

from __future__ import annotations

import contextlib

import pandas as pd

import dascore as dc
from dascore.constants import SpoolType
from dascore.io import FiberIO
from dascore.utils.hdf5 import (
    H5Reader,
    HDFPatchIndexManager,
    NodeError,
    PyTablesReader,
    PyTablesWriter,
)
from dascore.utils.misc import unbyte
from dascore.utils.patch import get_patch_names

from .utils import (
    _get_contents_from_patch_groups,
    _read_patch,
    _save_patch,
    _write_meta,
)


class DASDAEV1(FiberIO):
    """
    Provides IO support for the DASDAE format version 1.

    DASDAE format is loosely based on the Adaptable Seismic Data Format (ASDF)
    which uses hdf5. The hdf5 structure is the following:

    /root
    /root.attrs
        __format__ = "DASDAE"
        __DASDAE_version__ = '1'  # version str
    /root/waveforms/
        DAS__{net}__{sta}__{tag}__{start}__{end}
            data   # patch data array
            data.attrs
            _coords_{coord_name}  # each coordinate array is saved here
        DAS__{net}__{sta}__{tag}__{start}__{end}.attrs
            _attrs_{attr_nme}  # each patch attribute
            _dims  # a str of 'dim1, dim2, dim3'
    """

    name = "DASDAE"
    preferred_extensions = ("h5", "hdf5")
    version = "1"

    def write(self, spool: SpoolType, resource: PyTablesWriter, index=False, **kwargs):
        """
        Write a collection of patches to a DASDAE file.

        Parameters
        ----------
        spool:
            A collection of patches or a spool (same thing).
        resource
            The path to the file.
        index
            If True, create an index for the file. This slows down
            writing but will make reading/scanning the file more efficient.
            This is recommended for files with many patches and not recommended
            for files with few patches.
        """
        # write out patches
        _write_meta(resource, self.version)
        # get an iterable of patches and save them
        patches = [spool] if isinstance(spool, dc.Patch) else spool
        # create new node called waveforms, else suppress error if it
        # already exists.
        with contextlib.suppress(NodeError):
            resource.create_group(resource.root, "waveforms")
        waveforms = resource.get_node("/waveforms")
        # write new patches to file
        patch_names = get_patch_names(patches).values
        for patch, name in zip(patches, patch_names):
            _save_patch(patch, waveforms, resource, name)
        indexer = HDFPatchIndexManager(resource)
        if index or indexer.has_index:
            df = self._get_patch_summary(patches)
            indexer.write_update(df)

    def _get_patch_summary(self, patches) -> pd.DataFrame:
        """Get a patch summary to put into index."""
        df = (
            dc.scan_to_df(patches)
            .assign(
                path=lambda x: get_patch_names(x),
                file_format=self.name,
                file_version=self.version,
            )
            .dropna(subset=["time_min", "time_max", "distance_min", "distance_max"])
        )
        return df

    def get_format(self, resource: H5Reader, **kwargs) -> tuple[str, str] | bool:
        """Return the format from a dasdae file."""
        is_dasdae, version = False, ""  # NOQA
        attrs = resource.attrs
        file_format = unbyte(attrs.get("__format__", ""))
        if file_format != self.name:
            return False
        version = unbyte(attrs.get("__DASDAE_version__", ""))
        return file_format, version

    def read(self, resource: PyTablesReader, **kwargs) -> SpoolType:
        """Read a dascore file."""
        patches = []
        try:
            waveform_group = resource.root["/waveforms"]
        except (KeyError, IndexError):
            return dc.spool([])
        for patch_group in waveform_group:
            patches.append(_read_patch(patch_group, **kwargs))
        return dc.spool(patches)

    def scan(self, resource: PyTablesReader, **kwargs):
        """
        Get the patch info from the file.

        First we check if the file is using an index. If so, the index is
        returned. Otherwise, iterate over each waveform group and assemble
        content information.

        Parameters
        ----------
        resource
            A path to the file.
        """
        indexer = HDFPatchIndexManager(resource.filename)
        if indexer.has_index:
            # We need to change the path back to the file rather than internal
            # HDF5 path so it works with FileSpool and such.
            records = indexer.get_index().assign(path=str(resource)).to_dict("records")
            return [dc.PatchAttrs(**x) for x in records]
        else:
            file_format = self.name
            version = resource.root._v_attrs.__DASDAE_version__
            return _get_contents_from_patch_groups(resource, version, file_format)

    def index(self, path):
        """Index the dasdae file."""
        indexer = HDFPatchIndexManager(path)
        if not indexer.has_index:
            df = dc.scan_to_df(path)
            indexer.write_update(df)



"""IO module for reading simple h5 data."""

from __future__ import annotations

import dascore as dc
from dascore.constants import SpoolType
from dascore.io import FiberIO
from dascore.utils.hdf5 import H5Reader, PyTablesReader

from .utils import _get_attrs_coords_and_data, _is_h5simple, _maybe_trim_data


class H5Simple(FiberIO):
    """Support for bare-bones h5 format."""

    name = "H5Simple"
    preferred_extensions = ("hdf5", "h5")
    version = "1"

    def get_format(self, resource: H5Reader, **kwargs) -> tuple[str, str] | bool:
        """Determine if is simple h5 format."""
        if _is_h5simple(resource):
            return self.name, self.version
        return False

    def read(self, resource: PyTablesReader, snap=True, **kwargs) -> SpoolType:
        """
        Read a simple h5 file.

        Parameters
        ----------
        resource
            The open h5 object.
        snap
            If True, snap each coordinate to be evenly sampled.
        **kwargs
            Passed to filtering coordinates.
        """
        attrs, cm, data = _get_attrs_coords_and_data(resource, snap, self)
        new_cm, new_data = _maybe_trim_data(cm, data, kwargs)
        patch = dc.Patch(coords=new_cm, data=new_data[:], attrs=attrs)
        return dc.spool([patch])

    def scan(
        self, resource: PyTablesReader, snap=True, **kwargs
    ) -> list[dc.PatchAttrs]:
        """Get the attributes of a h5simple file."""
        attrs, cm, data = _get_attrs_coords_and_data(resource, snap, self)
        attrs["coords"] = cm.to_summary_dict()
        attrs["path"] = resource.filename
        return [dc.PatchAttrs(**attrs)]



"""
Base functionality for reading, writing, determining file formats, and scanning
Das Data.
"""

from __future__ import annotations

import inspect
import os.path
import warnings
from collections import defaultdict
from collections.abc import Generator
from functools import cache, cached_property, wraps
from importlib.metadata import entry_points
from pathlib import Path
from typing import Annotated, Literal, get_type_hints

import numpy as np
import pandas as pd
from pydantic import ConfigDict, Field, model_validator

import dascore as dc
from dascore.compat import Progress
from dascore.constants import (
    PROGRESS_LEVELS,
    VALID_DATA_CATEGORIES,
    VALID_DATA_TYPES,
    PatchType,
    SpoolType,
    max_lens,
    timeable_types,
)
from dascore.core.attrs import str_validator
from dascore.core.spool import DataFrameSpool
from dascore.exceptions import (
    InvalidFiberFileError,
    InvalidFiberIOError,
    MissingOptionalDependencyError,
    UnknownFiberFormatError,
)
from dascore.utils.io import IOResourceManager, get_handle_from_resource
from dascore.utils.mapping import FrozenDict
from dascore.utils.misc import _iter_filesystem, cached_method, iterate, warn_or_raise
from dascore.utils.models import (
    CommaSeparatedStr,
    DascoreBaseModel,
    DateTime64,
    TimeDelta64,
)
from dascore.utils.pd import _model_list_to_df
from dascore.utils.progress import track


class PatchFileSummary(DascoreBaseModel):
    """
    The necessary attributes for indexing a fiber file.

    A subset of [PatchAttributes](`dascore.core.attrs.PatchAttrs`).
    """

    model_config = ConfigDict(
        title="Patch File Summary",
        extra="ignore",
    )

    data_type: Annotated[Literal[VALID_DATA_TYPES], str_validator] = ""
    data_category: Annotated[Literal[VALID_DATA_CATEGORIES], str_validator] = ""
    instrument_id: str = Field("", max_length=max_lens["instrument_id"])
    experiment_id: str = Field("", max_length=max_lens["experiment_id"])
    tag: str = Field("", max_length=max_lens["tag"])
    station: str = Field("", max_length=max_lens["station"])
    network: str = Field("", max_length=max_lens["network"])
    dims: CommaSeparatedStr = Field("", max_length=max_lens["dims"])
    time_min: DateTime64 = np.datetime64("NaT")
    time_max: DateTime64 = np.datetime64("NaT")
    time_step: TimeDelta64 = np.timedelta64("NaT")
    # the attributes to index on
    file_version: str = ""
    file_format: str = ""
    path: str | Path = ""

    @property
    def dim_tuple(self):
        """Return a tuple of dimensions (eg ("time", "distance"))."""
        return tuple(self.dims.split(","))

    @model_validator(mode="before")
    @classmethod
    def translate_d_to_step(cls, data):
        """Translate d_time and d_distance to time_step, distance_step."""
        if isinstance(data, dict):
            for name in ["time", "distance"]:
                step_name, d_name = f"{name}_step", f"d_{name}"
                if step_name not in data and d_name in data:
                    data[step_name] = data.pop(d_name)
        return data

    def flat_dump(self):
        """Alias for dump, for compatibility with PatchAttrs.flat_dump."""
        return self.model_dump()


class _FiberIOManager:
    """
    A structure for intelligently storing, loading, and return FiberIO objects.

    This should only be used in conjunction with `FiberIO`.
    """

    def __init__(self, entry_point: str):
        self._entry_point = entry_point
        self._loaded_eps: set[str] = set()
        self._format_version = defaultdict(dict)
        self._extension_list = defaultdict(list)
        # This is a dict of {input_type: (fiberio_name, version)}
        self._fiber_io_by_input_type = defaultdict(set)
        self._fiber_io_name_ver = set()

    @cached_property
    def _eps(self):
        """
        Get the unloaded entry points registered to this domain into a dict of
        {name: ep}.
        """
        fiber_io_eps = entry_points(group="dascore.fiber_io")
        out = {x.name: x.load for x in fiber_io_eps}
        return pd.Series(out)

    @cached_property
    def known_formats(self):
        """Return names of known formats."""
        formats = self._eps.index.str.split("__").str[0]
        return set(formats) | set(self._format_version)

    @property
    def unloaded_formats(self):
        """Return names of known formats."""
        return sorted(self.known_formats - set(self._format_version))

    @cached_method
    def _get_fiber_io_by_input_type(self, input_type) -> set[FiberIO]:
        """Get a set of FiberIO instances that meet input type."""
        if input_type not in self._fiber_io_by_input_type:
            out = set()
            for input_set in self._fiber_io_by_input_type.values():
                out |= input_set
        else:
            out = self._fiber_io_by_input_type[input_type]
        return out

    @cache
    def _get_prioritized_list(self, input_type="file"):
        """Yield a prioritized list of fiber_ios."""
        # must load all plugins before getting list
        self.load_plugins()
        priority_fiber_ios = []
        second_class_fiber_ios = []
        for format_name in self.known_formats:
            unsorted = self._format_version[format_name]
            keys = sorted(unsorted, reverse=True)
            fiber_ios = [unsorted[key] for key in keys]
            priority_fiber_ios.append(fiber_ios[0])
            if len(fiber_ios) > 1:
                second_class_fiber_ios.extend(fiber_ios[1:])
        maybe_ios = priority_fiber_ios + second_class_fiber_ios
        # Now filter to input_type
        valid_fiberio_by_type = self._get_fiber_io_by_input_type(input_type)
        out = tuple(x for x in maybe_ios if x in valid_fiberio_by_type)
        # And return fiberIOs that much the input type.
        return out

    @cached_method
    def load_plugins(self, format: str | None = None):
        """Load plugin for specific format or ensure all formats are loaded."""
        if format is not None and format in self._format_version:
            return  # already loaded
        if not (unloaded := self.unloaded_formats):
            return
        formats = {format} if format is not None else unloaded
        # Load one, or all, formats
        for form in formats:
            for eps in self._eps.loc[self._eps.index.str.startswith(form)]:
                self.register_fiberio(eps()())
        # The selected format(s) should now be loaded
        assert set(formats).isdisjoint(self.unloaded_formats)

    def register_fiberio(self, fiberio: FiberIO):
        """Register a new fiber IO to manage."""
        forma, ver = fiberio.name.upper(), fiberio.version
        id_tuple = (forma, ver)
        if id_tuple in self._fiber_io_name_ver:
            return
        self._loaded_eps.add(fiberio.name)
        for ext in iter(fiberio.preferred_extensions):
            self._extension_list[ext].append(fiberio)
        self._format_version[forma][ver] = fiberio
        self._fiber_io_by_input_type[fiberio.input_type].add(fiberio)
        self._fiber_io_name_ver.add(id_tuple)

    @cached_method
    def get_fiberio(
        self,
        *,
        format: str | None = None,
        version: str | None = None,
        extension: str | None = None,
    ) -> FiberIO:
        """
        Return the most likely fiber_io for given inputs.

        If no such fiber_io exists, raise UnknownFiberFormat error.

        Parameters
        ----------
        format
            The format string indicating the format name
        version
            The version string of the format
        extension
            The extension of the file.
        """
        iterator = self.yield_fiberio(
            format=format,
            version=version,
            extension=extension,
        )
        for fiber_io in iterator:
            return fiber_io

    def yield_fiberio(
        self,
        format: str | None = None,
        version: str | None = None,
        extension: str | None = None,
        fiber_io_hint: dict[str, FiberIO] | None = None,
        input_type: str | None = None,
    ) -> Generator[FiberIO, None, None]:
        """
        Yields fiber IO object based on input priorities.

        The order is sorted in likelihood of the fiber_io being correct. For
        example, if file format is specified but file_version is not, all
        fiber_ios for the format will be yielded with the newest versions
        first in the list.

        If neither version nor format are specified but extension is all fiber_ios
        specifying the extension will be first in the list, sorted by format name
        and format version.

        If nothing is specified, all fiber_ios will be returned starting with
        the newest (the highest version) of each fiber_io, followed by older
        versions.

        Parameters
        ----------
        format
            The format string indicating the format name.
        version
            The version string of the format
        extension
            The extension of the file.
        fiber_io_hint
            If not None, a suspected fiber_io to use first. This is an
            optimization for file archives which tend to have many files of
            the same format.
        """
        fiber_io_hint = {} if fiber_io_hint is None else fiber_io_hint
        if version and not format:
            msg = "Providing only a version is not sufficient to determine format"
            raise UnknownFiberFormatError(msg)
        elif format is not None:
            self.load_plugins(format)
            yield from self._yield_format_version(format, version)
            return
        if input_type is not None and (out := fiber_io_hint.get(input_type)):
            yield out
        if extension is not None:
            yield from self._yield_extensions(extension, input_type)
        else:
            yield from self._get_prioritized_list(input_type)

    def _yield_format_version(self, format, version):
        """Yield file format/version prioritized fiber_ios."""
        assert isinstance(format, str), "Only works once format is known."
        format = format.upper()
        self.load_plugins(format)
        fiber_ios = self._format_version.get(format, None)
        # no format found
        if not fiber_ios:
            format_list = list(self.known_formats)
            msg = f"Unknown format {format}, " f"known formats are {format_list}"
            raise UnknownFiberFormatError(msg)
        # a version is specified
        if version:
            fiber_io = fiber_ios.get(version, None)
            if fiber_io is None:
                msg = (
                    f"Format {format} has no version: [{version}] "
                    f"known versions of this format are: {list(fiber_ios)}"
                )
                raise UnknownFiberFormatError(msg)
            yield fiber_io
            return
        # reverse sort fiber_ios and yield latest version first.
        for fiber_io in dict(sorted(fiber_ios.items(), reverse=True)).values():
            yield fiber_io
        return

    def _yield_extensions(self, extension, input_type=None):
        """Generator to get fiber_io prioritized by preferred extensions."""
        has_yielded = set()
        self.load_plugins()
        potential_fiberios = self._get_fiber_io_by_input_type(input_type)
        for fiber_io in self._extension_list[extension]:
            if fiber_io in potential_fiberios:
                yield fiber_io
            has_yielded.add(fiber_io)
        for fiber_io in self._get_prioritized_list(input_type):
            if fiber_io not in has_yielded:
                yield fiber_io

    def _get_format(
        self,
        path: str | Path | IOResourceManager,
        file_format: str | None = None,
        file_version: str | None = None,
        fiber_io_hint: dict[str, FiberIO] | None = None,
        **kwargs,
    ) -> tuple[str, str]:
        """
        Return the name of the format contained in the file and version number.

        See [`dascore.io.core.get_format`](`dascore.io.core.get_format`)
        for docs.
        """
        with IOResourceManager(path) as man:
            path = man.source
            if not os.path.exists(path):
                raise FileNotFoundError(f"{path} does not exist.")
            # get extension (str minus .)
            suffix = Path(path).suffix
            ext = suffix[1:] if suffix else None
            input_type = self._get_input_type_name(path)
            iterator = self.yield_fiberio(
                file_format,
                file_version,
                extension=ext,
                fiber_io_hint=fiber_io_hint,
                input_type=input_type,
            )
            for fiber_io in iterator:
                # We need to wrap this in try except to make it robust to what
                # may happen in each fiber_ios get_format method, many of which
                # may be third party code.
                func = fiber_io.get_format
                required_type = func._required_type
                func_input = None
                try:
                    # Get resource has to be in the try block because it can also
                    # raise, in which case the format doesn't belong.
                    func_input = man.get_resource(required_type)
                    format_version = func(func_input, _pre_cast=True)
                # For robustness, we need to catch everything else here.
                except Exception:
                    continue
                finally:
                    # If file handle-like seek back to 0 so it can be reused.
                    getattr(func_input, "seek", lambda x: None)(0)
                if format_version:
                    return format_version
            else:
                msg = f"Could not determine file format of {man.source}"
                raise UnknownFiberFormatError(msg)

    def _get_input_type_name(self, obj):
        """Get the name of the IO type."""
        # This effectively acts as a dispatch to determine which type of
        # FiberIO could possibly read the obj.
        out = "file"
        if isinstance(obj, str | Path) and (path := Path(obj)).exists():
            out = "directory" if path.is_dir() else "file"
        return out


# ------------- Protocol for File Format support


def _type_caster(func, sig, required_type, arg_name):
    """A decorator for casting types for arguments of cast ind."""
    fun_name = func.__name__

    # this is a subclass of a FiberIO subclass and its key methods
    # have already been wrapped. Just return.
    if getattr(func, "_type_caster_wrapped", False):
        return func

    @wraps(func)
    def _wrapper(*args, _pre_cast=False, **kwargs):
        """Wraps args but performs coercion to get proper stream."""
        # TODO look at replacing this with pydantic's type_guard thing.

        # this allows us to fast-track calls from generic functions
        if required_type is None or _pre_cast:
            return func(*args, **kwargs)
        bound = sig.bind(*args, **kwargs)
        new_kw = bound.arguments
        resource = new_kw.pop(arg_name)
        try:
            new_resource = get_handle_from_resource(resource, required_type)
            new_kw[arg_name] = new_resource
            # kwargs is included in bound arguments, need to re-attach
            new_kw.update(new_kw.pop("kwargs", {}))
            out = func(**new_kw)
        except Exception as e:  # get_format can't raise; must return false.
            if fun_name == "get_format":
                out = False
            else:
                raise e
        else:
            # if a new file handle was created we need to close it now. But it
            # shouldn't close any passed in, that should happen up the stack.
            if new_resource is not resource and hasattr(new_resource, "close"):
                new_resource.close()
        return out

    # attach the function and required type for later use
    _wrapper.func = func
    # subclasses of FIBERIO subclasses can wrap this twice, so we mark
    # it to avoid that scenario.
    _wrapper._type_caster_wrapped = True
    # also specify required type
    _wrapper._required_type = required_type

    return _wrapper


def _is_wrapped_func(func1, func2):
    """Small helper function to determine if func1 is func2, unwrapping decorators."""
    func = func1
    while hasattr(func, "func") or hasattr(func, "__func__"):
        func = getattr(func, "func", func)
        func = getattr(func, "__func__", func)
    return func is func2


class FiberIO:
    """
    An interface which adds support for a given filer format.

    This class should be subclassed when adding support for new formats.
    """

    name: str = ""
    version: str = ""
    preferred_extensions: tuple[str] = ()
    # Specifies if this fiber IO expects a directory or single file
    input_type: Literal["file", "directory"] = "file"

    manager = _FiberIOManager("dascore.fiber_io")

    # A dict of methods which should implement automatic type casting.
    # and the index of the parameter to type cast.
    _automatic_type_casters = FrozenDict(
        {
            "read": 1,
            "scan": 1,
            "write": 2,
            "get_format": 1,
        }
    )

    def read(self, resource, **kwargs) -> SpoolType:
        """
        Load data from a path.

        *kwargs should include support for selecting expected dimensions. For
        example, distance=(100, 200) would only read data with distance from
        100 to 200.
        """
        msg = f"FiberIO: {self.name} has no read method"
        raise NotImplementedError(msg)

    def scan(self, resource, **kwargs) -> list[dc.PatchAttrs]:
        """Returns a list of summary info for patches contained in file."""
        # default scan method reads in the file and returns required attributes
        # however, this can be very slow, so each parser should implement scan
        # when possible.
        try:
            spool = self.read(resource)
        except NotImplementedError:
            msg = f"FiberIO: {self.name} has no scan or read method"
            raise NotImplementedError(msg)
        out = []
        for pa in spool:
            new = pa.attrs.update(
                file_format=self.name,
                path=str(resource),
            )
            out.append(new)
        return out

    def write(self, spool: SpoolType, resource, **kwargs):
        """Write the spool to a resource (eg path, stream, etc.)."""
        msg = f"FiberIO: {self.name} has no write method"
        raise NotImplementedError(msg)

    def get_format(self, resource, **kwargs) -> tuple[str, str] | bool:
        """
        Return a tuple of (format_name, version_numbers).

        This should only work if path is the supported file format, otherwise
        raise UnknownFiberError or return False.
        """
        msg = f"FiberIO: {self.name} has no get_version method"
        raise NotImplementedError(msg)

    @property
    def implements_read(self) -> bool:
        """Returns True if the subclass implements its own scan method else False."""
        return not _is_wrapped_func(self.read, FiberIO.read)

    @property
    def implements_write(self) -> bool:
        """Returns True if the subclass implements its own scan method else False."""
        return not _is_wrapped_func(self.write, FiberIO.write)

    @property
    def implements_scan(self) -> bool:
        """Returns True if the subclass implements its own scan method else False."""
        return not _is_wrapped_func(self.scan, FiberIO.scan)

    @property
    def implements_get_format(self) -> bool:
        """Return True if the subclass implements its own get_format method."""
        return not _is_wrapped_func(self.get_format, FiberIO.get_format)

    @classmethod
    def get_supported_io_table(cls):
        """Make a table of all the supported formats and the methods."""
        # load all the plugins, so we know about all the FiberIO classes
        FiberIO.manager.load_plugins()
        out = []
        # iterate the dict _format_version_items,
        # which has the form {format_name: {version_str: FiberIO}}
        for format_name, version_dict in FiberIO.manager._format_version.items():
            for version_name, fiberio in version_dict.items():
                format_info = {
                    "name": format_name,
                    "version": version_name,
                    "scan": fiberio.implements_scan,
                    "get_format": fiberio.implements_get_format,
                    "read": fiberio.implements_read,
                    "write": fiberio.implements_write,
                }
                out.append(format_info)
        return pd.DataFrame(out)

    def _updated_after(self, resource, timestamp):
        """Determine if the resource was updated after specified mtime."""
        if not timestamp:
            return True
        return Path(resource).stat().st_mtime > timestamp

    def __hash__(self):
        """FiberIO instances should be uniquely defined by (format, version)."""
        return hash((self.name, self.version))

    def __init_subclass__(cls, **kwargs):
        """Hook for registering subclasses."""
        # check that the subclass is valid
        if not cls.name:
            msg = "You must specify the file format with the name field."
            raise InvalidFiberIOError(msg)
        # register fiber_io
        manager: _FiberIOManager = cls.__mro__[1].manager
        manager.register_fiberio(cls())
        # decorate methods for type-casting
        for name, param_ind in cls._automatic_type_casters.items():
            method = getattr(cls, name)
            sig = inspect.signature(method)
            arg_name = list(sig.parameters)[param_ind]
            required_type = get_type_hints(method).get(arg_name)
            method_wrapped = _type_caster(method, sig, required_type, arg_name)
            setattr(cls, name, method_wrapped)


def read(
    path: str | Path | IOResourceManager,
    file_format: str | None = None,
    file_version: str | None = None,
    time: tuple[timeable_types | None, timeable_types | None] | None = None,
    distance: tuple[float | None, float | None] | None = None,
    **kwargs,
) -> SpoolType:
    """
    Read a fiber file.

    For most cases, [`dascore.spool`](`dascore.spool`) is preferable to
    this function.

    Parameters
    ----------
    path
        A path to the file to read.
    file_format
        A string indicating the file format. If not provided dascore will
        try to estimate the format.
    file_version
        An optional string indicating the format version.
    time
        An optional tuple of time ranges.
    distance
        An optional tuple of distances.
    *kwargs
        All kwargs are passed to the format-specific read functions.

    Notes
    -----
    Unlike [`spool`](`dascore.spool`) this function reads the entire file
    into memory.

    Examples
    --------
    >>> import dascore as dc
    >>> from dascore.utils.downloader import fetch
    >>>
    >>> file_path = fetch("prodml_2.1.h5")
    >>>
    >>> patch = dc.read(file_path)
    """
    with IOResourceManager(path) as man:
        if not file_format or not file_version:
            file_format, file_version = get_format(
                man,
                file_format=file_format,
                file_version=file_version,
            )
        fiber_io = FiberIO.manager.get_fiberio(format=file_format, version=file_version)
        required_type = fiber_io.read._required_type
        path = man.get_resource(required_type)
        out = fiber_io.read(
            path,
            file_version=file_version,
            time=time,
            distance=distance,
            _pre_cast=True,
            **kwargs,
        )
        # if resource has a seek go back to 0 so this stream can be re-used.
        getattr(path, "seek", lambda x: None)(0)
        return out


def scan_to_df(
    path: Path | str | PatchType | SpoolType | IOResourceManager | pd.DataFrame,
    file_format: str | None = None,
    file_version: str | None = None,
    ext: str | None = None,
    timestamp: float | None = None,
    progress: PROGRESS_LEVELS = "standard",
    exclude=("history",),
) -> pd.DataFrame:
    """
    Scan a path, return a dataframe of contents.

    The columns of the dataframe depend on the attributes and coordinates
    found in the data files.

    Parameters
    ----------
    path
        The path to the to file to scan
    file_format
        Format of the file. If not provided DASCore will try to determine it.
    file_version
        The version string of the file.
    exclude
        A sequence of column names to exclude in the final dataframe.

    Examples
    --------
    >>> import dascore as dc
    >>> from dascore.utils.downloader import fetch
    >>>
    >>> file_path = fetch("prodml_2.1.h5")
    >>>
    >>> df = dc.scan_to_df(file_path)
    """
    if isinstance(path, pd.DataFrame):
        return path
    if isinstance(path, DataFrameSpool):
        return path.get_contents()
    info = scan(
        path=path,
        file_format=file_format,
        file_version=file_version,
        ext=ext,
        timestamp=timestamp,
        progress=progress,
    )
    df = _model_list_to_df(info, exclude=exclude)
    return df


def _iterate_scan_inputs(patch_source, ext, mtime, include_directories=True, **kwargs):
    """Yield scan candidates."""
    for el in iterate(patch_source):
        if isinstance(el, str | Path) and (path := Path(el)).exists():
            generator = _iter_filesystem(
                path, ext=ext, timestamp=mtime, include_directories=include_directories
            )
            yield from generator
        else:
            yield el


def _get_fiber_io_and_req_type(
    manager,
    file_format: str | None = None,
    file_version: str | None = None,
    fiber_io_hint=None,
):
    """
    Get the fiber IO for a patch source.

    Raises
    ------
    UnknownFileFormatError if no format is determinable from the
    patch_source

    """
    if not file_format or not file_version:
        file_format_, file_version_ = FiberIO.manager._get_format(
            path=manager,
            file_format=file_format,
            file_version=file_version,
            fiber_io_hint=fiber_io_hint,
        )
    else:
        # we need separate loop variables so this doesn't get assumed
        # to be the version/format in all subsequent values for the loop.
        file_format_, file_version_ = file_format, file_version
    fiber_io_hint = FiberIO.manager.get_fiberio(
        format=file_format_, version=file_version_
    )
    req_type = getattr(fiber_io_hint.scan, "_required_type", None)
    resource = manager.get_resource(req_type)
    # this will get the required resource type to pass to scan.
    return fiber_io_hint, resource


def _count_generator(generator):
    """Estimate the number of updates needed."""
    # TODO: This is a but sloppy, need to think of a better way to do
    # this to avoid double iteration.
    # First get total number of possible update-able files
    entity_count = 0
    for _ in generator:
        entity_count += 1
    return entity_count


def _handle_missing_optionals(outputs, optional_dep_dict):
    """
    Inform the user there are files that can be read but the proper
    dependencies are not installed.

    If there are other readable files that were found, raise a warning.
    Otherwise, raise a MissingOptionalDependencyError.
    """
    msg = (
        f"DASCore found files that can be read if additional packages are "
        f"installed. The needed packages and the found number of files are: "
        f"{dict(optional_dep_dict)}"
    )
    warn_or_raise(
        msg,
        exception=MissingOptionalDependencyError,
        warning=UserWarning,
        behavior="warn" if len(outputs) else "raise",
    )


def scan(
    path: Path | str | PatchType | SpoolType | IOResourceManager,
    file_format: str | None = None,
    file_version: str | None = None,
    ext: str | None = None,
    timestamp: float | None = None,
    progress: PROGRESS_LEVELS | Progress = "standard",
) -> list[dc.PatchAttrs]:
    """
    Scan a potential patch source, return a list of PatchAttrs.

    Parameters
    ----------
    path
        A resource containing Fiber data.
    file_format
        Format of the file. If not provided DASCore will try to determine it.
        Only applicable for path-like inputs.
    file_version
        Version of the file. If not provided DASCore will try to determine it.
        Only applicable for path-like inputs.
    ext : str or None
        The extensions to map.
    timestamp : int or float
        Time stamp indicating the minimum mtime.
    progress
        The type of progress bar to use. None disables progress bar and
        "basic" is best for low latency scenarios. Can also acceted a subclass
        of rich.progress.Progress.

    Returns
    -------
    A list of [`PatchAttrs`](`dascore.core.attrs.PatchAttrs`) or subclasses
    which may have extra fields.

    Examples
    --------
    >>> import dascore as dc
    >>> from dascore.utils.downloader import fetch
    >>>
    >>> file_path = fetch("prodml_2.1.h5")
    >>>
    >>> attr_list = dc.scan(file_path)

    See also [`iter_fs_contents`](`dascore.utils.misc.iter_fs_contents`)
    """
    out = []
    fiber_io_hint: dict[str, FiberIO] = {}
    # A dict for keeping track of missing optional dependencies.
    missing_optional_deps = defaultdict(lambda: 0)
    # Unfortunately, we have to iterate the scan candidates twice to get
    # an estimate for the progress bar length. Maybe there is a better way...
    _generator = _iterate_scan_inputs(
        path, ext=ext, mtime=timestamp, include_directories=False
    )
    length = _count_generator(_generator)
    generator = _iterate_scan_inputs(path, ext=ext, mtime=timestamp)
    # We want to avoid printing long object str reprs, so only print paths.
    resource_str = path if isinstance(path, str | Path) else ""
    tracker = track(
        generator,
        f"scan {resource_str}",
        progress=progress,
        length=length,
        min_length=20,
    )
    try:
        for patch_source in tracker:
            # just pull attrs from patch
            if isinstance(patch_source, dc.Patch):
                out.append(patch_source.attrs)
                continue
            with IOResourceManager(patch_source) as man:
                try:
                    fiber_io, resource = _get_fiber_io_and_req_type(
                        man,
                        file_format=file_format,
                        file_version=file_version,
                        fiber_io_hint=fiber_io_hint,
                    )
                except UnknownFiberFormatError:  # skip bad entities
                    continue
                # Cache this fiber io to given preferential treatment next
                # iteration. This speeds up the common case of many files
                # with the same format.
                fiber_io_hint[fiber_io.input_type] = fiber_io
                # Special handling of directory FiberIOs.
                if fiber_io.input_type == "directory":
                    # Directory fiber_io should send skip signal back to generator
                    # so that no files/sub directories are scanned.
                    generator.send("skip")
                    if not fiber_io._updated_after(resource, timestamp):
                        continue
                    # Directory FiberIO may need to know the time after which
                    # contents should be returned.
                    source = fiber_io.scan(
                        resource, timestamp=timestamp, _pre_cast=True
                    )
                else:
                    try:
                        source = fiber_io.scan(resource, _pre_cast=True)
                    # This happens if the file is corrupt see #346.
                    except (OSError, InvalidFiberFileError, ValueError, TypeError):
                        warnings.warn(f"Failed to scan {resource}", UserWarning)
                        continue
                    except MissingOptionalDependencyError as ex:
                        missing_optional_deps[ex.msg.split(" ")[0]] += 1
                        continue
                for attr in source:
                    out.append(dc.PatchAttrs.from_dict(attr))
    # Ensure ctl + c exists scan.
    except KeyboardInterrupt:
        getattr(progress, "stop", lambda: None)()
        raise
    if missing_optional_deps:
        _handle_missing_optionals(out, missing_optional_deps)
    return out


def get_format(
    path: str | Path | IOResourceManager,
    file_format: str | None = None,
    file_version: str | None = None,
    fiber_io_hint: dict[str, FiberIO] | None = None,
    **kwargs,
) -> tuple[str, str]:
    """
    Return the name of the format contained in the file and version number.

    Parameters
    ----------
    path
        The path to the file.
    file_format
        The known file format.
    file_version
        The known file version.
    fiber_io_hint
        A dict of {input_type: fiber_io}. This is an optimization
        which assumes the last used fiberio (for a given input type)
        is likely to be the next one.

    Returns
    -------
    A tuple of (file_format_name, version) both as strings.

    Raises
    ------
    dascore.exceptions.UnknownFiberFormat - Could not determine the fiber format.

    Examples
    --------
    >>> import dascore as dc
    >>> from dascore.utils.downloader import fetch
    >>>
    >>> file_path = fetch("prodml_2.1.h5")
    >>>
    >>> file_format, file_version = dc.get_format(file_path)
    """
    out = FiberIO.manager._get_format(
        path, file_format, file_version, fiber_io_hint, **kwargs
    )
    return out


def write(
    patch_or_spool,
    path: str | Path,
    file_format: str,
    file_version: str | None = None,
    **kwargs,
) -> Path:
    """
    Write a Patch or Spool to disk.

    Parameters
    ----------
    path
        The path to the file.
    file_format
        The string indicating the format to write.
    file_version
        Optionally specify the version of the file, else use the latest
        version for the format.

    Raises
    ------
    [`UnkownFiberFormatError`](`dascore.exceptions.UnknownFiberFormatError`)
        - Could not determine the fiber format.

    Examples
    --------
    >>> from pathlib import Path
    >>> import dascore as dc
    >>>
    >>> patch = dc.get_example_patch()
    >>> path = Path("output.h5")
    >>> _ = dc.write(patch, path, "dasdae")
    >>>
    >>> assert path.exists()
    >>> path.unlink()
    """
    fiber_io = FiberIO.manager.get_fiberio(format=file_format, version=file_version)
    if not isinstance(patch_or_spool, dc.BaseSpool):
        patch_or_spool = dc.spool([patch_or_spool])
    with IOResourceManager(path) as man:
        func = fiber_io.write
        required_type = func._required_type
        resource = man.get_resource(required_type)
        func(patch_or_spool, resource, _pre_cast=True, **kwargs)
    return path



"""Tests for Fourier transforms."""

from __future__ import annotations

import numpy as np
import pytest

from dascore.transform.fft import rfft
from dascore.units import get_quantity
from dascore.utils.misc import suppress_warnings


class TestRfft:
    """Tests for the real fourier transform."""

    @pytest.fixture(scope="class")
    def rfft_patch(self, random_patch):
        """Return the random patched transformed along time w/ rrft."""
        with suppress_warnings(DeprecationWarning):
            out = rfft(random_patch, dim="time")
        return out

    def test_dims(self, rfft_patch):
        """Ensure ft of original axis shows up in dimensions."""
        dims = rfft_patch.dims
        start_freq = [x.startswith("ft_") for x in dims]
        assert any(start_freq)

    def test_abs_rrft(self, rfft_patch):
        """Ensure abs works with rfft to get amplitude spectra."""
        out = rfft_patch.abs()
        assert np.allclose(out.data, np.abs(rfft_patch.data))

    def test_time_coord_units(self, random_patch, rfft_patch):
        """Ensure time label units have been correctly set."""
        units1 = random_patch.coords.coord_map["time"].units
        units2 = rfft_patch.coords.coord_map["ft_time"].units
        assert get_quantity(units1) == 1 / get_quantity(units2)

    def test_data_units(self, random_patch):
        """Ensure data units have been updated."""
        patch = random_patch.update_attrs(data_units="m/s")
        with suppress_warnings(DeprecationWarning):
            fft_patch = patch.rfft("time")
        dunits1 = get_quantity(patch.attrs.data_units)
        dunits2 = get_quantity(fft_patch.attrs.data_units)
        assert dunits2 == dunits1 * get_quantity("second")



"""IO module for reading Terra15 DAS data."""

from __future__ import annotations

import dascore as dc
from dascore.constants import timeable_types
from dascore.io import FiberIO
from dascore.utils.hdf5 import H5Reader

from .utils import (
    _get_terra15_version_str,
    _get_version_data_node,
    _read_terra15,
    _scan_terra15,
)


class Terra15FormatterV4(FiberIO):
    """Support for Terra15 data format, version 4."""

    name = "TERRA15"
    preferred_extensions = ("hdf5", "h5")
    version = "4"

    def get_format(self, resource: H5Reader, **kwargs) -> tuple[str, str] | bool:
        """
        Return True if file contains terra15 version 2 data else False.

        Parameters
        ----------
        resource
            A path to the file which may contain terra15 data.
        """
        version_str = _get_terra15_version_str(resource)
        if version_str:
            return (self.name, version_str)

    def scan(self, resource: H5Reader, **kwargs) -> list[dc.PatchAttrs]:
        """Scan a terra15 v2 file, return summary information."""
        version, data_node = _get_version_data_node(resource)
        extras = {
            "path": resource.filename,
            "file_format": self.name,
            "file_version": str(version),
        }
        return _scan_terra15(resource, data_node, extras)

    def read(
        self,
        resource: H5Reader,
        time: tuple[timeable_types, timeable_types] | None = None,
        distance: tuple[float, float] | None = None,
        snap_dims: bool = True,
        **kwargs,
    ) -> dc.BaseSpool:
        """
        Read a terra15 file.

        Parameters
        ----------
        resource
            The path to the file.
        time
            A tuple for filtering time.
        distance
            A tuple for filtering distance.
        snap_dims
            If True, ensure the coordinates are evenly sampled monotonic.
            This will cause some loss in precision but it is usually
            negligible.
        """
        patch = _read_terra15(resource, time, distance, snap_dims=snap_dims)
        return dc.spool(patch)


class Terra15FormatterV5(Terra15FormatterV4):
    """Support for Terra15 data format, version 5."""

    version = "5"


class Terra15FormatterV6(Terra15FormatterV4):
    """Support for Terra15 data format, version 5."""

    version = "6"



"""Utilities for chunking dataframes."""

from __future__ import annotations

from collections.abc import Collection
from functools import reduce

import numpy
import numpy as np
import pandas as pd

from dascore.constants import attr_conflict_description, numeric_types, timeable_types
from dascore.exceptions import ChunkError, CoordMergeError, ParameterError
from dascore.utils.docs import compose_docstring
from dascore.utils.misc import get_middle_value
from dascore.utils.pd import (
    _instructions_modified,
    _remove_overlaps,
    get_column_names_from_dim,
    get_dim_names_from_columns,
    get_interval_columns,
)
from dascore.utils.time import (
    is_datetime64,
    is_timedelta64,
    to_datetime64,
    to_timedelta64,
)


def get_intervals(
    start,
    stop,
    length,
    overlap=None,
    step=None,
    keep_partials=False,
):
    """
    Create a range of values with optional overlaps.

    Parameters
    ----------
    start
        The start of the interval.
    stop
        The end of the interval.
    length
        The length of the segments.
    overlap
        The overlap of the start of each interval with the end
        of the previous interval.
    step
        If not None, subtract step (the sampling interval) from the end
        values so that the intervals do not overlap by one sample.
    keep_partials
        If True, keep the segments which are smaller than chunksize.

    Returns
    -------
    A 2D array where first column is start and second column is end.
    """
    # when length is null just use entire length
    if pd.isnull(length):
        out = np.asarray([start, stop])
        if is_datetime64(start):
            out = to_datetime64(out)
        return np.atleast_2d(out)

    if is_datetime64(start):
        # need to ensure we have numpy datetimes, not pandas
        start, stop = to_datetime64(start), to_datetime64(stop)
        length = to_timedelta64(length)
    # get variable and perform checks
    overlap = length * 0 if not overlap else overlap
    step = length * 0 if pd.isnull(step) else step
    # Check for errors
    if overlap > length:
        msg = "Cant chunk when overlap is greater than chunk size"
        raise ParameterError(msg)
    # If the step is known, we need to account for it in the total duration
    # See 474.
    _raw_duration = stop - start
    duration = _raw_duration + step if step is not None else _raw_duration
    if duration < length and not keep_partials:
        msg = "Cant chunk when data interval is less than chunk size. "
        raise ChunkError(msg)
    # reference with no overlap
    new_step = length - overlap
    reference = np.arange(start, stop + new_step, step=new_step)
    # Since we just add to get stop values we need to remove anything
    # that is within a sample of stopping value (otherwise that segment
    # will have no data).
    reference = reference[(reference + step) <= stop]
    # we subtract step to avoid overlaps in segments. This can mean segments
    # are ~ one sample shorter than those requested.
    ends = reference + length - step
    starts = reference
    # trim end to not surpass stop
    bad_ends = ends > stop
    if bad_ends.any():
        if not keep_partials:
            ends_filt = ends <= stop
            ends, starts = ends[ends_filt], starts[ends_filt]
        else:
            ends[bad_ends] = stop
    return np.stack([starts, ends]).T


@compose_docstring(attr_conflict=attr_conflict_description)
class ChunkManager:
    """
    A class for managing the chunking of data defined in a dataframe.

    The chunk manager handles both splitting and joining of contiguous,
    or near-contiguous, blocks of data.

    Parameters
    ----------
    overlap
        The amount of overlap between each segment, starting with the end of
        first row. Negative values can be used for inducing gaps.
    group_columns
        A sequence of column names which should be used for sorting groups.
    keep_partial
        If True, keep segments which are shorter than chunk size (at end of
        contiguous blocks)
    tolerance
        The upper limit of a gap to tolerate in terms of the sampling
        along the desired dimension. E.G., the default value means entities
        with gaps <= 1.5 * {name}_step will be merged.
    conflict
        {attr_conflict}
    **kawrgs
        kwargs specify the column along which to chunk. The key specifies the
        column along which to chunk, typically, `time` or `distance`, and the
        value specifies the chunk size. A value of None means to chunk on all
        available data (e.g. merge all data).

    Notes
    -----
    This class is used internally by `dc.BaseSpool.chunk`.
    """

    def __init__(
        self,
        overlap: timeable_types | numeric_types | None = None,
        group_columns: Collection[str] | None = None,
        keep_partial=False,
        snap_coords=True,
        tolerance=1.5,
        conflict="raise",
        **kwargs,
    ):
        self._overlap = overlap
        self._group_columns = group_columns
        self._keep_partials = keep_partial
        self._snap_coords = snap_coords
        self._tolerance = tolerance
        self._name, self._value = self._validate_kwargs(kwargs)
        self._attr_conflict = conflict
        self._validate_chunker()

    def _validate_kwargs(self, kwargs):
        """Ensure kwargs is len one and has a valid."""
        if not len(kwargs) == 1:
            msg = (
                f"Chunking only supported along one dimension. You passed "
                f"kwargs: {kwargs}"
            )
            raise ParameterError(msg)
        ((key, value),) = kwargs.items()
        value = None if value is ... else value
        return key, value

    def _validate_chunker(self):
        """Ensure selected parameters are compatible."""
        # chunker is used for merging
        if pd.isnull(self._value):
            if self._keep_partials or self._overlap:
                msg = (
                    "When chunk value is None (ie Chunker is used for merging) "
                    "both _keep_partials and self._overlap must not be selected."
                )
                raise ParameterError(msg)
            return
        # ensure chunk values are greater than 0
        zero = to_timedelta64(0) if is_timedelta64(self._value) else 0
        if self._value <= zero:
            msg = "Chunk value must be greater than 0."
            raise ParameterError(msg)

    def _get_continuity_group_number(self, start, stop, step) -> pd.Series:
        """Return a series of ints indicating continuity group."""
        # start by sorting according to start time
        arg_ser = start.argsort()
        args = arg_ser.index[arg_ser.values]
        start_sorted, stop_sorted, step_sorted = start[args], stop[args], step[args]
        # next get cummax of endtimes and detect gaps
        stop_cum_max = stop_sorted.cummax()
        end_markers = stop_cum_max.shift() + step_sorted * self._tolerance
        has_gap = start_sorted > end_markers
        group_num = has_gap.astype(np.int64).cumsum()
        return group_num[start.index]

    def _get_sampling_group_num(self, step, tolerance=0.05) -> pd.Series:
        """
        Because sampling can be off a little, this adds some tolerance for
        how sampling affects groups.

        Tolerance affects how close samples have to be in order to count as
        the same. 5% is used here.
        """
        col = step.values
        sort_args = np.argsort(col)
        sorted_col = col[sort_args]
        roll_forward = np.roll(sorted_col, shift=1)
        diff = (sorted_col - roll_forward) / sorted_col
        out_of_threshold = diff > tolerance
        group_number = numpy.cumsum(out_of_threshold)
        # undo sorting
        out = pd.Series(group_number[np.argsort(sort_args)], index=step.index)
        return out

    def _get_duration_overlap(self, duration, start, step, overlap=None):
        """Get duration and overlap from kwargs."""
        overlap = overlap if overlap is not None else self._overlap
        # cast step to time delta if start is datetime
        if is_datetime64(start):
            step = to_timedelta64(step)
            overlap = to_timedelta64(overlap)
        if pd.isnull(overlap):
            overlap = np.asarray([0], dtype=step.dtype)[0]
        return duration, overlap

    def _create_df(self, df, name, start_stop, gnum):
        """Reconstruct the dataframe."""
        cols = f"{name}_min", f"{name}_max"
        out = pd.DataFrame(start_stop, columns=list(cols))
        out[f"{name}_step"] = get_middle_value(df[f"{name}_step"].values)
        merger = df.drop(columns=out.columns)
        # get dims to determine which columns are still compared. Some test
        # dfs don't have dims though, so it should still work without dims col.
        dims = set(df.iloc[0].get("dims", "").split(","))
        # We exclude private columns for considering if merge can happen.
        for col in set(x for x in merger.columns if not x.startswith("_")):
            prefix = col.split("_")[0]
            # If we have specified to ignore or remove conflicting attrs
            # we don't need to check them here, but we do still check dims.
            if self._attr_conflict != "raise" and prefix not in dims:
                continue
            vals = merger[col].unique()
            if len(vals) > 1:
                msg = (
                    f"Cannot merge on dim {self._name} because all values for "
                    f"{col} are not equal. Consider using the `conflict` "
                    f"argument to loosen this restriction."
                )
                raise CoordMergeError(msg)

            assert len(vals) == 1, "Haven't yet implemented non-homogenous merging"
            out[col] = vals[0]
        # add the group number for getting instruction df later
        out["_group"] = gnum
        return out

    def _get_chunk_overlap_inds(self, src1, src2, chu1, chu2):
        """Get an index mapping from source to chunk."""
        chunk_starts = np.searchsorted(src1, chu1, side="right") - 1
        chunk_ends = np.searchsorted(src2, chu2, side="left")
        # Ensure no chunks run off the end of the source.
        assert np.all(chunk_ends < len(src1)), "Invalid chunk range found"
        # add 1 to end so it is an exclusive end range
        return np.stack([chunk_starts, chunk_ends + 1], axis=1)

    def _get_source_and_chunk_inds(self, chunk2src_inds, s_index, c_index):
        """Get ndarrays of chunk index, source index."""
        # get indices for sorted arrays
        source_inds_ = np.concatenate(
            [np.arange(x[0], x[1], dtype=np.int64) for x in chunk2src_inds]
        )
        chunk_inds_ = np.concatenate(
            [
                np.ones((x[1] - x[0]), dtype=np.int64) * num
                for num, x in enumerate(chunk2src_inds)
            ]
        )
        # use pandex index to map back to actual indices
        source_inds = s_index.values[source_inds_]
        chunk_inds = c_index.values[chunk_inds_]
        out = {
            "source_sorted": source_inds_,
            "source": source_inds,
            "chunk_sorted": chunk_inds_,
            "chunk": chunk_inds,
        }
        return out

    def _get_instructions(self, sub_source, sub_chunk):
        """Get source mapping to chunk."""
        min_name, max_name = f"{self._name}_min", f"{self._name}_max"
        # sort inputs based on start of range, as long as we don't reset index
        # we should be ok.
        sub_source = sub_source.sort_values(min_name)
        sub_chunk = sub_chunk.sort_values(min_name)
        # need to make sure we don't have overlaps in source df. This implicitly
        # handles merging.
        sub_source = _remove_overlaps(sub_source, self._name)
        src1, src2, src_step = get_interval_columns(sub_source, self._name, arrays=True)
        chu1, chu2, chu_step = get_interval_columns(sub_chunk, self._name, arrays=True)
        dims = get_dim_names_from_columns(sub_source)
        cols2keep = get_column_names_from_dim(dims)
        # next get index range for which chunk times belong to.
        chunk2src_inds = self._get_chunk_overlap_inds(src1, src2, chu1, chu2)
        # total length of source to chunk mapping
        inds = self._get_source_and_chunk_inds(
            chunk2src_inds,
            sub_source.index,
            sub_chunk.index,
        )
        source_inds, chunk_inds = inds["source_sorted"], inds["chunk_sorted"]
        # get potential start/stop times.
        starts = np.stack([src1[source_inds], chu1[chunk_inds]], axis=1)
        ends = np.stack([src2[source_inds], chu2[chunk_inds]], axis=1)
        end_values = np.min(ends, axis=1)
        start_values = np.max(starts, axis=1)
        data_dict = {
            min_name: start_values,
            max_name: end_values,
            "source_index": inds["source"],
            "current_index": inds["chunk"],
        }
        out = pd.DataFrame(data_dict)
        # populate the rest of the columns needed in instruction df.
        for col in cols2keep:
            if col in out.columns:
                continue
            out[col] = sub_source[col].values[source_inds]
        out = out.sort_index()
        out["_modified"] = _instructions_modified(out, sub_source)
        return out

    def get_instruction_df(self, source_df, chunked_df):
        """
        Get a dataframe connecting the chunked dataframe to its origin.

        This is used to connect source data to desired data after chunking
        operation.

        Parameters
        ----------
        source_df
            The dataframe before chunking
        chunked_df
            The chunked dataframe (output of `chunk` method)
        """
        # the group column should exist and the chunked groups should be subset
        # of the source groups
        assert "_group" in source_df.columns and "_group" in chunked_df.columns
        chunked_groups = set(chunked_df["_group"])
        if not chunked_groups:
            return pd.DataFrame(columns=[*list(source_df.columns), "_modified"])
        # chunk groups should be a subset of source groups
        assert chunked_groups.issubset(set(source_df["_group"]))
        # iterate each group and create instruction df
        out = []
        for group in chunked_groups:
            sub_source = source_df[source_df["_group"] == group]
            sub_chunk = chunked_df[chunked_df["_group"] == group]
            out.append(self._get_instructions(sub_source, sub_chunk))
        df = pd.concat(out, axis=0).reset_index(drop=True).set_index("source_index")
        return df

    def _get_col_group(self, df, cont_g):
        """Get group columns based on common columns."""
        cols = list(self._group_columns or [])
        columns = [x for x in cols if x in df.columns]
        col_g = cont_g * 0 if not columns else df.groupby(columns).ngroup()
        return col_g

    def _get_group(self, df, start, stop, step):
        """
        Get the group designation for df. This accounts for both time intervals
        being consistent and group columns matching.
        """
        cont_g = self._get_continuity_group_number(start, stop, step)
        samp_g = self._get_sampling_group_num(step)
        col_g = self._get_col_group(df, cont_g)
        group_series = [x.astype(str) for x in [samp_g, col_g, cont_g]]
        group = reduce(lambda x, y: x + "_" + y, group_series)
        return group

    def _get_group_dfs(self, group, dur, overlap, group_mins, group_maxs, df, step):
        """Get the new dataframe for a given group."""
        out = []
        for gnum in group.unique():
            g_start, g_stop = group_mins[gnum], group_maxs[gnum]
            current_df = df.loc[group[group == gnum].index]
            # reconstruct DF
            try:
                new_start_stop = get_intervals(
                    g_start,
                    g_stop,
                    dur,
                    overlap=overlap,
                    step=step.iloc[0],
                    keep_partials=self._keep_partials,
                )
            except ChunkError:  # this chunk is too short, skip.
                continue
            # create the newly chunked dataframe
            sub_new_df = self._create_df(current_df, self._name, new_start_stop, gnum)
            out.append(sub_new_df)
        return out

    def _filter_nan_dfs(self, df, start, stop):
        """Filter NaN out of dataframe if they occur in start/stop."""

    def chunk(
        self,
        df: pd.DataFrame,
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Chunk a dataframe into new contiguous segments.

        The dataframe must have column names {key}_max, {key}_min, and {key}_step
        where {key} is the key used in the kwargs.

        Parameters
        ----------
        df
            Input dataframe to chunk.

        Returns
        -------
        A tuple of the original dataframe with added column '_group' and an
        output dataframe with column '_group'. The _group column is used
        to link the two dataframes together.
        """
        if df.empty:  # empy df, do nothing
            return df.assign(_group=None), df.assign(_group=None)
        # get series of start/stop along requested dimension
        start, stop, step = get_interval_columns(df, self._name)
        # Filter out any NaN in start or stop.
        keep = ~(pd.isnull(start) | pd.isnull(stop))
        df, start, stop, step = df[keep], start[keep], stop[keep], step[keep]
        if df.empty:  # Need to check again since NaN can wipe out df.
            return df.assign(_group=None), df.assign(_group=None)
        dur, overlap = self._get_duration_overlap(self._value, start, step)
        # get group numbers
        group = self._get_group(df, start, stop, step)
        # get max, min for each group and expand
        group_mins = start.groupby(group).min()
        group_maxs = stop.groupby(group).max()
        # split/group dataframe into new chunks by iterating over each group.
        out = self._get_group_dfs(group, dur, overlap, group_mins, group_maxs, df, step)
        if not len(out):
            msg = "Could not chunk. No segments with sufficient length found."
            raise ChunkError(msg)
        out = pd.concat(out, axis=0).reset_index(drop=True)
        return df.assign(_group=group), out



"""
Module for applying just in time compilation to speed up functions.
"""

from __future__ import annotations

import warnings
from functools import wraps


class _DummyNumba:
    """A simple class for acting like numba when numba is not installed."""

    # This isn't intended to make everything work, only basic stuff needed
    # for tests to pass when numba isn't installed.
    def prange(self, count):
        """A mock of numba's prange."""
        yield from range(count)


def maybe_numba_jit(required=False, _missing_numba=False, **compiler_kwargs):
    """
    Use numba to apply JIT compilation to the decorated function.

    Parameters
    ----------
    required
        If True an ImportError is raised if the wrapped function is called
        and the compiler module is not installed. If False, issue a warning.
    _missing_numba
        If true, simulate missing the numba package. Only used for testing.
    **compiler_kwargs
        Keyword arguments passed to the compiler function.

    Examples
    --------
    >>> import numpy as np
    >>> from dascore.utils.jit import maybe_numba_jit
    >>>
    >>> # JIT function for fast numba compilation
    >>> @maybe_numba_jit(nopython=True, nogil=True)
    ... def jit_func(array):
    ...     ...
    >>>
    >>> # You can use the numba module inside the function without importing
    >>> # it because numba is added to the functions globals by the decorator.
    >>> # To get linters to ignore this seemingly undefined variable use #noqa
    >>> @maybe_numba_jit
    ... def jit_func(array):
    ...     for a in numba.prange(len(array)):  # noqa
    ...         pass
    ...     return array
    >>>

    >>> out = jit_func(np.array([1,2,3]))

    Notes
    -----
    - By design, the warning/exception is only raised when calling the wrapped
      function. Users who don't use the function should be unaffected.
    - After the jit the original function can be accessed via the `func`
      attribute. This is useful to testing in python mode.
    """
    # This happens when the decorator is used without parens; use default values
    if callable(required):
        return maybe_numba_jit()(required)

    has_numba = True
    try:
        import numba

        if _missing_numba:
            raise ImportError("Simulating missing numba.")
    except ImportError:
        numba = _DummyNumba()
        has_numba = False

    def _wraper(func):
        # Add numba to the functions global namespace so that it can be used
        # within the function without having to import it.
        globs = getattr(func, "__globals__", {})
        globs["numba"] = numba

        if not has_numba:

            @wraps(func)
            def decorated(*args, **kwargs):
                if required:
                    msg = (
                        f"{func.__name__} requires python module "
                        f"numba but it is not installed. "
                    )
                    raise ImportError(msg)
                else:
                    msg = (
                        f"{func.__name__} can be compiled to improve performance. "
                        f"Please install numba to enable JIT."
                    )
                    warnings.warn(msg, UserWarning)
                return func(*args, **kwargs)

            out_func = decorated
        else:
            out_func = numba.jit(**compiler_kwargs)(func)
        out_func.func = func  # make original func accessible via .func
        return out_func

    return _wraper



"""Core module for wave format."""

from __future__ import annotations

from pathlib import Path

import numpy as np
from scipy.io.wavfile import write

from dascore.constants import ONE_SECOND, SpoolType
from dascore.io.core import FiberIO
from dascore.utils.patch import check_patch_coords


class WavIO(FiberIO):
    """IO support for wav (audio) format."""

    name = "WAV"

    def write(
        self, spool: SpoolType, resource: str | Path, resample_frequency=None, **kwargs
    ):
        """
        Write the contents of the patch to one or more wav files.

        Parameters
        ----------
        resource
            If a path that ends with .wav, write all non-time channels
            to a single file. If not, assume the path is a directory and
            write each non-time channel to its own wav file.
        resample_frequency
            A resample frequency in Hz. If None, do not perform resampling.
            Often DAS has non-int sampling rates, so the default resampling
            rate is usually safe.

        Notes
        -----
            - The sampling rate in the wav format must be an int, so the
            patch's sampling rate is cast to an integer before writing wav file.
            This may cause some distortion (but it isn't likely to be noticeable).

            - The array data type is converted to np.float32 before writing. This
            requires values to be between (-1, 1) so the data are detrended
            and normalized before writing.

            - If a single wavefile is specified with the path argument, and
            the output the patch has more than one len along the non-time
            dimension, a multi-channel wavefile is created. There may be some
            players that do not support multi-channel wavefiles.

            - If using VLC, often it won't play the file unless the sampling
            rate is 44100, in this case just set resample_frequency=44100 to
            see if this fixes the issue.
        """
        resource = Path(resource)
        assert len(spool) == 1, "Only single patch spools can be written to wav"
        patch = spool[0]
        # write a single wav file, maybe multi-channeled.
        data, sr = self._get_wav_data(patch, resample_frequency)
        if resource.name.endswith(".wav"):
            write(filename=str(resource), rate=int(sr), data=data)
        else:  # write data to directory, one file for each non-time
            resource.mkdir(exist_ok=True, parents=True)
            non_time_set = set(patch.dims) - {"time"}
            non_time_name = next(iter(non_time_set))
            non_time = patch.coords.get_array(non_time_name)
            for ind, val in enumerate(non_time):
                sub_data = np.take(data, ind, axis=1)
                sub_path = resource / f"{non_time_name}_{val}.wav"
                write(filename=str(sub_path), rate=int(sr), data=sub_data)

    @staticmethod
    def _get_wav_data(patch, resample):
        """Pre-condition patch data for writing. Return array and sample rate."""
        # Ensure we have a 2D patch which has a time dimension.
        check_patch_coords(patch, ("time",))
        assert len(patch.dims) == 2, "only 2D patches supported for this function."
        time = patch.get_coord("time").step

        # handle resampling and normalization
        pat = patch.transpose("time", ...)
        if resample is not None:
            pat = pat.resample(time=1 / resample)
        # normalize and detrend
        pat = pat.detrend("time", "linear").normalize("time", norm="max")
        data = pat.data
        sample_rate = resample or np.round(ONE_SECOND / time)
        return data.astype(np.float32), int(sample_rate)



"""IO module for reading binary raw format DAS data."""

from __future__ import annotations

from pathlib import Path
from xml.etree.ElementTree import ParseError

import numpy as np
from pydantic import ValidationError

import dascore as dc
from dascore.io import FiberIO
from dascore.utils.models import UTF8Str

from .utils import _load_patches, _paths_to_attrs, _read_xml_metadata


class BinaryPatchAttrs(dc.PatchAttrs):
    """Patch attrs for Binary."""

    pulse_width_ns: float = np.nan
    gauge_length: float = np.nan
    instrument_id: UTF8Str = ""
    distance_units: UTF8Str = ""
    zone_name: UTF8Str = ""


class XMLBinaryV1(FiberIO):
    """Support for binary data format with xml metadata."""

    name = "XMLBinary"
    version = "1"
    input_type = "directory"

    _metadata_name = "metadata.xml"
    # File extension for data files.
    _data_extension = ".raw"

    def scan(self, resource, timestamp=None, **kwargs) -> list[dc.PatchAttrs]:
        """Scan the contents of the directory."""
        path = Path(resource)
        metadata = _read_xml_metadata(path / self._metadata_name)
        data_files = list(path.glob(f"*{self._data_extension}"))
        extra_attrs = {
            "file_version": self.version,
            "file_format": self.name,
        }
        # Need to update time
        attrs = _paths_to_attrs(
            data_files,
            metadata,
            timestamp=timestamp,
            attr_cls=BinaryPatchAttrs,
            extra_attrs=extra_attrs,
        )
        return attrs

    def read(self, resource, time=None, distance=None, **kwargs) -> dc.BaseSpool:
        """
        Load data from the directory structure.

        Parameters
        ----------
        resource
            A directory, path to the index file, or path to a data file.
        time
            Parameters for filtering by time.
        distance
            Parameters for filtering by distance.
        **kwargs
            Extra keyword arguments are ignored.
        """
        path = Path(resource)
        base_path = path if path.is_dir() else path.parent
        meta_data = _read_xml_metadata(base_path / self._metadata_name)
        if path.is_dir():
            path = list(path.glob(f"*{self._data_extension}"))
        # Determine if this is a single file or all of them.
        patches = _load_patches(
            path,
            meta_data,
            time=time,
            distance=distance,
            attr_cls=BinaryPatchAttrs,
        )
        return dc.spool(patches)

    def get_format(self, resource, **kwargs) -> tuple[str, str] | bool:
        """Determine if directory is an XML Binary type."""
        path = Path(resource)
        index_path = path / self._metadata_name
        if not index_path.exists():
            return False
        try:
            _ = _read_xml_metadata(index_path)
        except (ParseError, TypeError, IndexError, ValidationError):
            return False
        return self.name, self.version



"""
Tests for attr utilities.
"""

from __future__ import annotations

import pytest

import dascore as dc
from dascore import PatchAttrs
from dascore.exceptions import AttributeMergeError
from dascore.utils.attrs import (
    combine_patch_attrs,
    decompose_attrs,
    separate_coord_info,
)
from dascore.utils.downloader import fetcher


class TestMergeAttrs:
    """Tests for merging patch attrs."""

    def test_empty(self):
        """Empty PatchAttrs should work in all cases."""
        pa1, pa2 = PatchAttrs(), PatchAttrs()
        assert isinstance(combine_patch_attrs([pa1, pa2]), PatchAttrs)
        out = combine_patch_attrs([pa1, pa2], drop_attrs="history")
        assert isinstance(out, PatchAttrs)

    def test_empty_with_coord_raises(self):
        """Attributes which don't have specified coord should raise."""
        pa1, pa2 = PatchAttrs(), PatchAttrs()
        match = "Failed to combine"
        with pytest.raises(AttributeMergeError, match=match):
            combine_patch_attrs([pa1, pa2], "time")
        with pytest.raises(AttributeMergeError, match=match):
            combine_patch_attrs([pa1, pa2], "distance")

    def test_simple_merge(self):
        """Happy path, simple merge."""
        pa1 = PatchAttrs(distance_min=1, distance_max=10)
        pa2 = PatchAttrs(distance_min=11, distance_max=20)
        merged = combine_patch_attrs([pa1, pa2], "distance")
        # the names of attrs should be identical before/after merge
        assert set(dict(merged)) == set(dict(pa1))
        assert merged.distance_max == pa2.distance_max
        assert merged.distance_min == pa1.distance_min

    def test_drop(self):
        """Ensure drop_attrs does its job."""
        pa1 = PatchAttrs(history=["a", "b"])
        pa2 = PatchAttrs()
        msg = "the following non-dim attrs are not equal"
        with pytest.raises(AttributeMergeError, match=msg):
            combine_patch_attrs([pa1, pa2])
        out = combine_patch_attrs([pa1, pa2], drop_attrs="history")
        assert isinstance(out, PatchAttrs)

    def test_conflicts(self):
        """Ensure when non-dim fields aren't equal merge raises."""
        pa1 = PatchAttrs(tag="bob", another=2, same=42)
        pa2 = PatchAttrs(tag="bob", another=2, same=42)
        pa3 = PatchAttrs(another=1, same=42, different=10)
        msg = "the following non-dim attrs are not equal"
        with pytest.raises(AttributeMergeError, match=msg):
            combine_patch_attrs([pa1, pa2, pa3])

    def test_missing_coordinate(self):
        """When one model has a missing coord it should just get dropped."""
        pa1 = PatchAttrs(bob_min=10, bob_max=12)
        pa2 = PatchAttrs(bob_min=13)
        out = combine_patch_attrs([pa1, pa2], coord_name="bob")
        assert out == pa1

    def test_drop_conflicts(self, random_patch):
        """Ensure when non-dim fields aren't equal, but are defined, they drop."""
        pa1 = random_patch.attrs.update(tag="bill", station="UU")
        pa2 = random_patch.attrs.update(station="TA", tag="bob")
        out = combine_patch_attrs([pa1, pa2], coord_name="time", conflicts="drop")
        defaults = PatchAttrs()
        assert isinstance(out, PatchAttrs)
        assert out.tag == defaults.tag
        assert out.station == defaults.station

    def test_keep_disjoint_values(self, random_patch):
        """Ensure when disjoint values should be kept they are."""
        random_attrs = random_patch.attrs
        attrs1 = random_attrs.update(jazz_hands=1984)
        out = combine_patch_attrs([attrs1, random_attrs], conflicts="keep_first")
        assert out.jazz_hands == 1984

    def test_unequal_raises(self, random_patch):
        """When attrs have unequal coords it should raise an error."""
        attr1 = random_patch.attrs
        attr2 = attr1.update(distance_units="miles")
        match = "Cant merge patch attrs"
        with pytest.raises(AttributeMergeError, match=match):
            combine_patch_attrs([attr1, attr2], coord_name="distance")


class TestDecomposeAttrs:
    """Tests for decomposing attributes."""

    @pytest.fixture(scope="class")
    def scanned_attrs(self):
        """Scan all the attrs in the dascore cache, return them."""
        return dc.scan(fetcher.path)

    @pytest.fixture(scope="class")
    def decomposed_attrs(self, scanned_attrs):
        """Decompose attrs into dicts of lists."""
        return decompose_attrs(scanned_attrs)

    def test_only_datetime(self, decomposed_attrs):
        """Ensure all times are datetime64."""
        time_dtypes = decomposed_attrs["dims"]["time"]
        assert set(time_dtypes) == {"datetime64"}


class TestSeparateCoordInfo:
    """Tests for separating coord info from attr dict."""

    def test_empty(self):
        """Empty args should return emtpy dicts."""
        out1, out2 = separate_coord_info(None)
        assert out1 == out2 == {}

    def test_meets_reqs(self):
        """Simple case for filtering out required attrs."""
        input_dict = {"coords": {"time": {"min": 10}}}
        coords, attrs = separate_coord_info(input_dict)
        assert coords == input_dict["coords"]

    def test_dict_of_coord_info(self, random_patch):
        """Passing in a dictionary of coord info should work."""
        coord_dict = random_patch.coords.to_summary_dict()
        dims = random_patch.dims
        coords, attrs = separate_coord_info(coord_dict, dims=dims)
        assert coords == coord_dict



"""Tests for performing aggregations."""

from __future__ import annotations

import numpy as np
import numpy.testing as npt
import pytest

import dascore as dc
from dascore.exceptions import ParameterError
from dascore.proc.aggregate import _AGG_FUNCS
from dascore.utils.misc import broadcast_for_index


class TestBasicAggregations:
    """Sanity checks for basic aggregations."""

    def test_first(self, random_patch):
        """Ensure aggregations can occur."""
        out = random_patch.aggregate(dim="distance", method="first")
        assert out.ndim == random_patch.ndim
        axis = random_patch.dims.index("distance")
        inds = broadcast_for_index(len(random_patch.data.shape), axis, 0)
        assert np.allclose(random_patch.data[inds].flatten(), out.data.flatten())

    def test_last(self, random_patch):
        """Ensure aggregations can occur."""
        out = random_patch.aggregate(dim="time", method="last")
        axis = random_patch.dims.index("time")
        inds = broadcast_for_index(len(random_patch.data.shape), axis, -1)
        assert np.allclose(random_patch.data[inds].flatten(), out.data.flatten())

    def test_no_dim(self, random_patch):
        """Ensure no dimension argument behaves like numpy."""
        out = random_patch.aggregate(method="mean")
        expected = np.mean(random_patch.data, keepdims=True)
        npt.assert_allclose(out.data, expected, rtol=1e-8, atol=0)

    def test_dtype_of_coord_unchanged(self, random_patch):
        """The dtype of the coord should not change."""
        out = random_patch.aggregate(dim="time", method="median")
        coord_new = out.get_coord("time")
        coord_old = random_patch.get_coord("time")
        assert coord_old.dtype == coord_new.dtype

    @pytest.mark.parametrize("method", list(_AGG_FUNCS))
    def test_named_aggregations(self, random_patch, method):
        """Simply run the named aggregations."""
        patch = getattr(random_patch, method)(dim="distance")
        assert isinstance(patch, dc.Patch)

    def test_dim_reduce_squeeze(self, random_patch):
        """Ensure the old dimension can be squeezed out."""
        out = random_patch.aggregate(dim="time", method="mean", dim_reduce="squeeze")
        assert "time" not in out.dims

    def test_dim_reduce_mean(self, random_patch):
        """Ensure the mean value can be left on the coord."""
        out = random_patch.aggregate(dim="time", method="mean", dim_reduce="mean")
        new_time = out.get_coord("time")
        assert len(new_time) == 1

    def test_dim_reduce_mean_time_delta(self, random_patch):
        """Ensure the mean value can be left on the coord."""
        time = random_patch.get_coord("time")
        dt = dc.to_timedelta64(time.values)
        patch = random_patch.update_coords(time=dt)
        out = patch.aggregate(dim="time", method="mean", dim_reduce="mean")
        new_time = out.get_coord("time")
        assert len(new_time) == 1

    def test_invalid_dim_reduce(self, random_patch):
        """Ensure an invalid dim_reduce argument raises."""
        msg = "dim_reduce must be"
        with pytest.raises(ParameterError, match=msg):
            random_patch.aggregate(dim="time", dim_reduce="invalid")

    def test_dim_reduce_first(self, random_patch):
        """Ensure first takes the first value"""
        out = random_patch.aggregate(dim="time", method="mean", dim_reduce="first")
        new_time = out.get_coord("time")
        assert len(new_time) == 1
        assert new_time[0] == out.get_array("time")[0]

    def test_dim_reduce_distance(self, random_patch):
        """Ensure non-time dims also work."""
        out = random_patch.aggregate(dim="distance", method="mean", dim_reduce=np.var)
        assert "distance" in out.dims
        expected = np.var(random_patch.get_array("distance"))
        assert out.get_coord("distance").values == expected


class TestApplyOperators:
    """Ensure aggregated patches can be used as operators for arithmetic."""

    def test_complete_reduction(self, random_patch):
        """Ensure a patch with complete reduction works."""
        agg = random_patch.min(None)
        assert np.all(agg.data == np.min(random_patch.data))
        # Ensure broadcasting works with reduced data
        out = random_patch - agg
        assert isinstance(out, dc.Patch)
        assert np.allclose(random_patch.data - agg, out.data)

    def test_single_reduction(self, random_patch):
        """Ensure a single patch reduced also works with broadcasting."""
        agg = random_patch.first("time")
        out1 = random_patch + agg
        assert isinstance(out1, dc.Patch)



"""
dascore Filtering module.

Much of this code was inspired by ObsPy's filtering module created by:
Tobias Megies, Moritz Beyreuther, Yannik Behr
"""

from __future__ import annotations

import sys
from collections.abc import Sequence

import numpy as np
import pandas as pd
from scipy import ndimage
from scipy.ndimage import gaussian_filter as np_gauss
from scipy.ndimage import median_filter as nd_median_filter
from scipy.signal import filtfilt, iirfilter, iirnotch, sosfilt, sosfiltfilt, zpk2sos
from scipy.signal import savgol_filter as np_savgol_filter

import dascore as dc
from dascore.constants import PatchType, samples_arg_description
from dascore.exceptions import FilterValueError, ParameterError, UnitError
from dascore.units import (
    convert_units,
    get_filter_units,
    get_inverted_quant,
    invert_quantity,
    quant_sequence_to_quant_array,
)
from dascore.utils.docs import compose_docstring
from dascore.utils.misc import (
    broadcast_for_index,
    check_filter_kwargs,
    check_filter_range,
)
from dascore.utils.patch import (
    get_dim_axis_value,
    get_dim_sampling_rate,
    patch_function,
)
from dascore.utils.time import to_float


def _check_sobel_args(dim, mode, cval):
    """Check Sobel filter kwargs and return."""
    mode_options = {
        "reflect",
        "constant",
        "nearest",
        "mirror",
        "wrap",
        "grid-constant",
        "grid-mirror",
        "grid-wrap",
    }
    if not isinstance(dim, str):
        msg = "dim parameter should be a string."
        raise FilterValueError(msg)
    if not isinstance(mode, str):
        msg = "mode parameter should be a string."
        raise FilterValueError(msg)
    if not isinstance(cval, float | int):
        msg = "cval parameter should be a float or an int."
        raise FilterValueError(msg)
    if mode not in mode_options:
        msg = f"The valid values for modes are {mode_options}."
        raise FilterValueError(msg)

    return dim, mode, cval


def _get_sos(sr, filt_min, filt_max, corners):
    """Get second order sections from sampling rate and filter bounds."""
    nyquist = 0.5 * sr
    low = None if pd.isnull(filt_min) else filt_min / nyquist
    high = None if pd.isnull(filt_max) else filt_max / nyquist
    check_filter_range(nyquist, low, high, filt_min, filt_max)

    if (low is not None) and (high is not None):  # apply bandpass
        z, p, k = iirfilter(
            corners, [low, high], btype="band", ftype="butter", output="zpk"
        )
    elif low is not None:
        z, p, k = iirfilter(
            corners, low, btype="highpass", ftype="butter", output="zpk"
        )
    else:
        assert high is not None
        z, p, k = iirfilter(
            corners, high, btype="lowpass", ftype="butter", output="zpk"
        )
    return zpk2sos(z, p, k)


@patch_function()
def pass_filter(patch: PatchType, corners=4, zerophase=True, **kwargs) -> PatchType:
    """
    Apply a Butterworth pass filter (bandpass, highpass, or lowpass).

    Parameters
    ----------
    corners
        The number of corners for the filter. Default is 4.
    zerophase
        If True, apply the filter twice.
    **kwargs
        Used to specify the dimension and frequency, wavelength, or equivalent
        limits.

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()

    >>>  # Apply bandpass filter along time axis from 1 to 100 Hz
    >>> bandpassed = pa.pass_filter(time=(1, 100))

    >>>  # Apply lowpass filter along distance axis for wavelengths less than 100m
    >>> lowpassed = pa.pass_filter(distance=(None, 1/100))
    >>> # Note that None and ... both indicate open intervals
    >>> assert pa.pass_filter(time=(None, 90)) == pa.pass_filter(time=(..., 90))

    >>> # Optionally, units can be specified for a more expressive API.
    >>> from dascore.units import m, ft, s, Hz
    >>> # Filter from 1 Hz to 10 Hz in time dimension
    >>> lp_units = pa.pass_filter(time=(1 * Hz, 10 * Hz))
    >>> # Filter wavelengths 50m to 100m
    >>> bp_m = pa.pass_filter(distance=(50 * m, 100 * m))
    >>> # filter wavelengths less than 200 ft
    >>> lp_ft = pa.pass_filter(distance=(200 * ft, ...))
    """
    dim, (arg1, arg2) = check_filter_kwargs(kwargs)
    axis = patch.dims.index(dim)
    coord_units = patch.coords.coord_map[dim].units
    filt_min, filt_max = get_filter_units(arg1, arg2, to_unit=coord_units, dim=dim)
    sr = get_dim_sampling_rate(patch, dim)
    # get nyquist and low/high in terms of nyquist
    sos = _get_sos(sr, filt_min, filt_max, corners)
    if zerophase:
        out = sosfiltfilt(sos, patch.data, axis=axis)
    else:
        out = sosfilt(sos, patch.data, axis=axis)
    return dc.Patch(data=out, coords=patch.coords, attrs=patch.attrs, dims=patch.dims)


@patch_function()
def sobel_filter(patch: PatchType, dim: str, mode="reflect", cval=0.0) -> PatchType:
    """
    Apply a Sobel filter.

    Parameters
    ----------
    dim
        The dimension along which to apply
    mode
        Determines how the input array is extended when the filter
        overlaps with a border.
    cval
        Fill value when mode="constant".

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()

    >>>  # 1. Apply Sobel filter using the default parameter values.
    >>> sobel_default = pa.sobel_filter(dim='time', mode='reflect', cval=0.0)

    >>>  # 2. Apply Sobel filter with arbitrary parameter values.
    >>> sobel_arbitrary = pa.sobel_filter(dim='time', mode='constant', cval=1)

    >>> # 3. Apply Sobel filter along both axes
    >>> sobel_time_space = pa.sobel_filter('time').sobel_filter('distance')
    """
    dim, mode, cval = _check_sobel_args(dim, mode, cval)
    axis = patch.dims.index(dim)
    out = ndimage.sobel(patch.data, axis=axis, mode=mode, cval=cval)
    return dc.Patch(data=out, coords=patch.coords, attrs=patch.attrs, dims=patch.dims)


def _create_size_and_axes(patch, kwargs, samples):
    """
    Return a tuple of (size) and (axes).

    Note: size will always have the same size as the patch, but
    1s will be used if axis is not used.
    """
    dimfo = get_dim_axis_value(patch, kwargs=kwargs, allow_multiple=True)
    axes = [x.axis for x in dimfo]
    size = [1] * len(patch.dims)
    for dim, axis, value in dimfo:
        coord = patch.get_coord(dim)
        window = coord.get_sample_count(value, samples=samples)
        size[axis] = window
    return tuple(size), tuple(axes)


@patch_function()
@compose_docstring(sample_explination=samples_arg_description)
def median_filter(
    patch: PatchType, samples=False, mode="reflect", cval=0.0, **kwargs
) -> PatchType:
    """
    Apply 2-D median filter.

    Parameters
    ----------
    patch
        The patch to filter
    samples
        {sample_explination}
    mode
        The mode for handling edges.
    cval
        The constant value for when mode == constant.
    **kwargs
        Used to specify the shape of the median filter in each dimension.
        See examples for more info.

    Examples
    --------
    >>> import dascore
    >>> from dascore.units import m, s
    >>> pa = dascore.get_example_patch()

    >>>  # 1. Apply median filter only over time dimension with 0.10 sec window
    >>> filtered_pa_1 = pa.median_filter(time=0.1)

    >>>  # 2. Apply median filter over both time and distance
    >>>  # using a 0.1 second time window and 2 m distance window
    >>> filtered_pa_2 = pa.median_filter(time=0.1 * s, distance=2 * m)

    >>>  # 3. Apply median filter with 3 time samples and 4 distance samples
    >>> filtered_pa = pa.median_filter(
    ...     time=3, distance=4, samples=True,
    ... )

    Notes
    -----
    See scipy.ndimage.median_filter for more info on implementation
    and arguments.

    Values specified with kwargs should be small, for example < 10 samples
    otherwise this can take a long time and use lots of memory.
    """
    size, _ = _create_size_and_axes(patch, kwargs, samples)
    new_data = nd_median_filter(patch.data, size=size, mode=mode, cval=cval)
    return patch.update(data=new_data)


@patch_function()
def notch_filter(patch: PatchType, q, **kwargs) -> PatchType:
    """
    Apply a second-order IIR notch digital filter on patch's data.

    A notch filter is a band-stop filter with a narrow bandwidth (high quality factor).
    It rejects a narrow frequency band and leaves the rest of the spectrum
    little changed.

    Parameters
    ----------
    patch
        The patch to filter
    q
        Quality factor (float). A higher Q value means a narrower notch,
        which targets the specific frequency more precisely.
        A lower Q results in a wider notch, meaning a broader range of
        frequencies around the specific frequency will be attenuated.
    **kwargs
        Used to specify the dimension(s) and associated frequency and/or wavelength
        (or equivalent values) for the filter.

    See Also
    --------
    [scipy.signal.iirnotch](https://docs.scipy.org/doc/scipy/reference
    /generated/scipy.signal.iirnotch.html).

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()

    >>>  # Apply a notch filter along time axis to remove 60 Hz
    >>> filtered = pa.notch_filter(time=60, q=30)

    >>>  # Apply a notch filter along distance axis to remove 5 m wavelength
    >>> filtered = pa.notch_filter(distance=0.2, q=30)

    >>>  # Apply a notch filter along both time and distance axes
    >>> filtered = pa.notch_filter(time=60, distance=0.2, q=30)

    >>> # Optionally, units can be specified for a more expressive API.
    >>> from dascore.units import m, ft, s, Hz
    >>> # Apply a notch filter along time axis to remove 60 Hz
    >>> filtered = pa.notch_filter(time=60 * Hz, q=30)
    >>> # Apply a notch filter along distance axis to remove 5 m wavelength
    >>> filtered = pa.notch_filter(distance=5 * m, q=30)
    """
    data = patch.data
    dinfo = get_dim_axis_value(patch, kwargs=kwargs, allow_multiple=True)
    for dim, axis, value in dinfo:
        coord = patch.get_coord(dim)
        # Invert units if needed
        if isinstance(value, dc.units.Quantity) and coord.units is not None:
            value, _ = get_inverted_quant(value, coord.units)
        # Check valid parameters
        w0 = to_float(value)
        sr = get_dim_sampling_rate(patch, dim)
        nyquist = 0.5 * sr
        if w0 > nyquist:
            msg = f"possible filter values are in [0, {nyquist}] you passed {w0}"
            raise FilterValueError(msg)
        b, a = iirnotch(w0, Q=q, fs=sr)
        data = filtfilt(b, a, data, axis=axis)
    return dc.Patch(data=data, coords=patch.coords, attrs=patch.attrs, dims=patch.dims)


@patch_function()
@compose_docstring(sample_explination=samples_arg_description)
def savgol_filter(
    patch: PatchType, polyorder, samples=False, mode="interp", cval=0.0, **kwargs
) -> PatchType:
    """
    Applies Savgol filter along spenfied dimensions.

    The filter will be applied over each selected dimension sequentially.

    Parameters
    ----------
    patch
        The patch to filter
    polyorder
        Order of polynomial
    samples
        If True samples are specified
        If False coordinate of dimension
    mode
        The mode for handling edges.
    cval
        The constant value for when mode == constant.
    **kwargs
        Used to specify the shape of the savgol filter in each dimension.

    Notes
    -----
    See scipy.signal.savgol_filter for more info on implementation
    and arguments.

    Examples
    --------
    >>> import dascore
    >>> from dascore.units import m, s
    >>> pa = dascore.get_example_patch()
    >>>
    >>> # Apply second order polynomial Savgol filter
    >>> # over time dimension with 0.10 sec window.
    >>> filtered_pa_1 = pa.savgol_filter(polyorder=2, time=0.1)
    >>>
    >>> # Apply Savgol filter over distance dimension using a 5 sample
    >>> # distance window.
    >>> filtered_pa_2 = pa.savgol_filter(distance=5, samples=True, polyorder=2)
    >>>
    >>> # Combine distance and time filter
    >>> filtered_pa_3 = pa.savgol_filter(distance=10, time=0.1, polyorder=4)
    """
    data = patch.data
    size, axes = _create_size_and_axes(patch, kwargs, samples)
    for ax in axes:
        data = np_savgol_filter(
            x=patch.data,
            window_length=size[ax],
            polyorder=polyorder,
            mode=mode,
            cval=cval,
            axis=ax,
        )
    return patch.update(data=data)


@patch_function()
@compose_docstring(sample_explination=samples_arg_description)
def gaussian_filter(
    patch: PatchType, samples=False, mode="reflect", cval=0.0, truncate=4.0, **kwargs
) -> PatchType:
    """
    Applies a Gaussian filter along specified dimensions.

    Parameters
    ----------
    patch
        The patch to filter
    samples
        If True samples are specified
        If False coordinate of dimension
    mode
        The mode for handling edges.
    cval
        The constant value for when mode == constant.
    truncate
        Truncate the filter kernel length to this many standard deviations.
    **kwargs
        Used to specify the sigma value (standard deviation) for desired
        dimensions.

    Examples
    --------
    >>> import dascore
    >>> from dascore.units import m, s
    >>> pa = dascore.get_example_patch()
    >>>
    >>> # Apply Gaussian smoothing along time axis.
    >>> pa_1 = pa.gaussian_filter(time=0.1)
    >>>
    >>> # Apply Gaussian filter over distance dimension
    >>> # using a 3 sample standard deviation.
    >>> pa_2 = pa.gaussian_filter(samples=True, distance=3)
    >>>
    >>> # Apply filter to time and distance axis.
    >>> pa_3 = pa.gaussian_filter(time=0.1, distance=3)

    Notes
    -----
    See scipy.ndimage.gaussian_filter for more info on implementation
    and arguments.
    """
    size, axes = _create_size_and_axes(patch, kwargs, samples)
    used_size = tuple(size[x] for x in axes)
    data = np_gauss(
        input=patch.data,
        sigma=used_size,
        axes=axes,
        mode=mode,
        cval=cval,
        truncate=truncate,
    )
    return patch.update(data=data)


@patch_function()
@compose_docstring(sample_explination=samples_arg_description)
def slope_filter(
    patch: PatchType,
    filt: Sequence[float],
    dims: tuple[str, str] = ("distance", "time"),
    directional: bool = False,
    notch: bool = False,
) -> PatchType:
    """
    Filter the patch over certain slopes in the 2D Fourier domain.

    Most commonly this used as an F-K (frequency wavenumber)
    filter to attenuate energy with specified apparent velocities.

    Parameters
    ----------
    patch
        The patch to filter.
    filt
        A length 4 array of the form [va, vb, vc, vd]. If notch is False,
        the filter selects the apparent velocites between 'vb' and 'vc'
        with tapering boundaries from 'va' to 'vb' and from 'vc' to 'vd'.
    dims
        The dimensions used to determine slope. The first dim is in the
        numerator and the second in the denominator. (eg distance, time)
        represents a velocity since distance/time has units of |L|/|T|
        (commonly m/s).
    directional
        If True, the filter should be considered direction. That is to say,
        the sign of the values in `filt` indicate the direction (towards or
        away) with increasing coordinate values.
        This can be used for up-down/left-right separation, assuming a
        near-linear fiber layoyt.
    notch
        If True, the filter represents a notch, meaning the slopes
        specified by the inner `filt` parameterse are attenuated rather
        than those outside of them.

    Examples
    --------
    >>> # Example 1: Compare slope filtered patch to Non-filtered.
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>>
    >>> import dascore as dc
    >>> from dascore.units import Hz
    >>>
    >>>
    >>> # Apply taper function and bandpass filter along time axis from 1 to 500 Hz
    >>> patch = (
    ...     dc.get_example_patch('example_event_1')
    ...     .set_units(distance='m', time='s')
    ... 	.taper(time=0.05)
    ... 	.pass_filter(time=(1*Hz, 500*Hz))
    ... )
    >>> filt = np.array([2e3,2.2e3,8e3,2e4])
    >>> # Apply fk filter
    >>> patch_filtered = patch.slope_filter(
    ...     filt=filt,
    ... 	directional=False,
    ...     notch=False
    ... )
    >>> # Plot results
    >>> fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8))
    >>> ax1 = patch.viz.waterfall(ax=ax1, scale=0.5)
    >>> _ = ax1.set_title('Raw')
    >>> ax2 = patch_filtered.viz.waterfall(ax=ax2, scale=0.5)
    >>> _ = ax2.set_title('Filtered')
    >>>
    >>> # Example 2: Notch filter
    >>> patch_filtered = patch.slope_filter(filt=filt, notch=True)
    >>>
    >>> # Example 3: specify units
    >>> filt = np.array([2e3,2.2e3,8e3,2e4]) * dc.get_unit("m/s")
    >>> patch_filtered = patch.slope_filter(filt=filt)

    The [fk recipe](`docs/recipes/fk.qmd`) provides addtional examples.
    """

    def _check_inputs(patch, filt, dims):
        """Ensure inputs are valid."""
        sorted_filt = np.all(filt[:-1] <= filt[1:])
        if not (sorted_filt and len(filt) == 4):
            msg = f"filt must be a sorted length 4 sequence. You passed {filt}"
            raise ParameterError(msg)
        if missing := set(dims) - set(patch.coords.coord_map):
            msg = f"Cant apply slope filter. {missing} are missing from patch."
            raise ParameterError(msg)

    def _get_taper_mask(filt, slope, notch):
        """Get a mask for applying taper and attenuation."""
        fac = np.where(
            (slope >= filt[0]) & (slope <= filt[1]),
            1.0 - np.sin(0.5 * np.pi * (slope - filt[0]) / (filt[1] - filt[0])),
            1.0,
        )
        fac = np.where((slope >= filt[1]) & (slope <= filt[2]), 0.0, fac)
        fac = np.where(
            (slope >= filt[2]) & (slope <= filt[3]),
            np.sin(0.5 * np.pi * (slope - filt[2]) / (filt[3] - filt[2])),
            fac,
        )
        fac = fac if notch else 1.0 - fac
        return fac

    def _get_slope_array(dft_patch, directional, freq_dims):
        """Get an array which specifies slope."""
        dim1, dim2 = freq_dims[-1], freq_dims[-2]
        dims = dft_patch.dims
        ndims = dft_patch.ndim
        coord1 = dft_patch.get_array(dim1)
        coord2 = dft_patch.get_array(dim2) + sys.float_info.epsilon
        # Need to add appropriate blank dims to keep overall shape of patch.
        ax1, ax2 = dims.index(dim1), dims.index(dim2)
        shape_1 = broadcast_for_index(ndims, ax1, value=slice(None), fill=None)
        shape_2 = broadcast_for_index(ndims, ax2, value=slice(None), fill=None)
        # Then just allow broadcasting to do its magic
        slope = coord1[shape_1] / coord2[shape_2]
        if not directional:
            slope = np.abs(slope)
        return slope

    def _maybe_transform_units(filt, dft_patch, freq_dims):
        """Handle units on filter."""
        # Hand the units/partial units in sequence.
        units = getattr(filt, "units", None)
        try:
            filt = np.array(filt)
        except ValueError:
            filt = quant_sequence_to_quant_array(filt)
        if units:
            filt = filt * dc.get_quantity(units)
        if not isinstance(filt, dc.units.Quantity):
            return filt
        array, units = filt.magnitude, filt.units
        coord_unit_1 = dft_patch.get_coord(freq_dims[-1]).units
        coord_unit_2 = dft_patch.get_coord(freq_dims[-2]).units
        if not (coord_unit_1 and coord_unit_2):
            msg = (
                f"Units of {units} specified in Patch.slope_filter, but units "
                f"are not defined for both specified dimensions: {dims}."
            )
            raise UnitError(msg)
        new_units = coord_unit_1 / coord_unit_2
        # Determine if we need to flip units.
        if new_units.dimensionality == (1 / units).dimensionality:
            array, units = np.sort(1 / array), invert_quantity(units)
        out = convert_units(array, new_units, units)
        return out

    _check_inputs(patch, filt, dims)
    freq_dims = tuple(f"ft_{x}" for x in dims)
    dft_patch = patch.dft.func(patch, dims)
    transformed = patch is not dft_patch

    slope = _get_slope_array(dft_patch, directional, freq_dims)
    filt = _maybe_transform_units(filt, dft_patch, freq_dims)

    mask = _get_taper_mask(filt, slope, notch)
    new_data = dft_patch.data * mask
    out = dft_patch.update(data=new_data)
    if transformed:
        out = out.idft().real()
    return out



"""Tests for chunking dataframes."""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

import dascore as dc
from dascore.exceptions import ParameterError
from dascore.utils.chunk import ChunkManager, get_intervals
from dascore.utils.time import to_timedelta64

STARTTIME = np.datetime64("2020-01-03")
ENDTIME = STARTTIME + np.timedelta64(60, "s")


@pytest.fixture()
def contiguous_df():
    """Create a contiguous dataframe with time and distance dimensions."""
    # get time, adjust starttime to be one time step after end time
    time = get_intervals(STARTTIME, ENDTIME, length=np.timedelta64(10, "s"))
    dt = np.timedelta64(10, "ms")
    df = pd.DataFrame(time, columns=["time_min", "time_max"])
    df["distance_min"], df["distance_max"] = 0, 10
    df["time_step"] = dt
    df["distance_step"] = 1
    return df


@pytest.fixture()
def contiguous_sr_spaced_df(contiguous_df):
    """Separate df by one sample rate."""
    sr = contiguous_df.loc[:, "time_step"]
    out = contiguous_df.copy()
    out["time_max"] = out["time_max"] - sr
    return out


@pytest.fixture()
def contiguous_df_two_stations(contiguous_df):
    """Create contiguous df with different stations."""
    # get time, adjust starttime to be one time step after end time
    df1 = contiguous_df.assign(station="sta1")
    df2 = contiguous_df.assign(station="sta2")
    return pd.concat([df1, df2], axis=0, ignore_index=True)


class TestGetIntervals:
    """Tests for generating intervals along some continuous dimension."""

    def test_numbers_no_overlap(self):
        """Ensure simple ints with no overlap work."""
        out = get_intervals(0, 10, 1)
        expected = np.stack([np.arange(10), np.arange(1, 11)]).T
        assert np.all(out == expected)

    def test_numbers_overlap(self):
        """Ensure numbers with start overlaps can also work."""
        start, stop, step = 0, 40, 10
        out = get_intervals(start, stop, step, overlap=1)
        expected = np.stack([np.array([0, 9, 18, 27]), np.array([10, 19, 28, 37])]).T

        # ensure the step size is the step specified for each interval
        assert np.allclose(out[:, 1] - out[:, 0], step)
        assert np.allclose(out, expected)

    def test_times_no_overlap(self):
        """Test using datetime64 and timedelta64."""
        start = np.datetime64("2017-01-03")
        step = np.timedelta64(1, "s")
        stop = start + 50 * step
        out = get_intervals(start, stop, step)
        assert np.all(out[:, 1] - out[:, 0] == step)
        assert out[-1, 1] == stop
        assert out[0, 0] == start

    def test_times_overlap(self):
        """Test time interval with overlap."""
        start = np.datetime64("2017-01-03")
        step = np.timedelta64(2, "s")
        overlap = np.timedelta64(1, "s")
        stop = start + 10 * step
        out = get_intervals(start, stop, step, overlap=overlap)
        assert np.all(out[:, 1] - out[:, 0] == step)
        assert out[-1, 1] == stop
        assert out[0, 0] == start


class TestBasicChunkDF:
    """Test basic DF chunking."""

    @pytest.fixture()
    def df_different_sample_rates(self, contiguous_df):
        """Tests for a df which does have overlaps but different sampling rates."""
        df1 = contiguous_df.copy()
        df2 = contiguous_df.copy()
        time_span = df1["time_max"].max() - df1["time_min"].min()
        df2["time_min"] += time_span
        df2["time_max"] += time_span
        df2["time_step"] = df1["time_step"] * 2
        out = pd.concat([df1, df2], axis=0).reset_index(drop=True)
        return out

    def test_rechunk_contiguous(self, contiguous_df):
        """Test rechunking with no gaps."""
        time_interval = (contiguous_df["time_max"] - contiguous_df["time_min"]).max()
        new_time_interval = time_interval / 2
        chunker = ChunkManager(time=new_time_interval)
        _, out = chunker.chunk(contiguous_df)
        assert len(out) == 2 * len(contiguous_df)
        time_step = out["time_step"].iloc[0]
        new_interval = (out["time_max"] - out["time_min"] + time_step).max()
        assert new_interval == new_time_interval

    def test_rechunk_contiguous_with_sr_separation(self, contiguous_sr_spaced_df):
        """Ensure it still works on data seperated by one sample."""
        df = contiguous_sr_spaced_df
        sr = df["time_step"]
        time_interval = (sr + df["time_max"] - df["time_min"]).max()
        new_time_interval = time_interval / 2
        chunker = ChunkManager(time=new_time_interval)
        _, out = chunker.chunk(df)
        assert len(out) == 2 * len(df)
        new_interval = (out["time_max"] - out["time_min"]).max()
        assert new_interval == (new_time_interval - sr.iloc[0])

    def test_rechunk_different_sr(self, df_different_sample_rates):
        """Ensure segments with different sample rates don't get combined."""
        df = df_different_sample_rates
        chunker = ChunkManager(overlap=None, time=23)
        _, out = chunker.chunk(df)
        dt = np.sort(np.unique(out["time_step"]))
        assert len(dt) == 2, "both dt should remain"
        # the second part of the df should start at the one minute mark
        df2 = out[out["time_step"] == dt[1]]
        time_min = df2.iloc[0]["time_min"]
        assert time_min.minute == 1

    def test_keep_leftovers(self, contiguous_df):
        """Ensure leftovers show up in df."""
        chunker = ChunkManager(overlap=None, keep_partial=True, time=28)
        _, out = chunker.chunk(contiguous_df)
        assert len(out) == 3
        assert out["time_max"].max() == contiguous_df["time_max"].max()

    def test_overlap(self, contiguous_df):
        """Ensure overlapping segments work."""
        over = to_timedelta64(10)
        chunker1 = ChunkManager(overlap=over, time=20)
        _, out = chunker1.chunk(contiguous_df)
        expected = over - contiguous_df["time_step"].iloc[0]
        olap = out.shift()["time_max"] - out["time_min"]
        assert np.all(pd.isnull(olap) | (olap == expected))
        # now ensure floats work for overlap param
        chunker2 = ChunkManager(overlap=10, time=20)
        _, out2 = chunker2.chunk(contiguous_df)
        assert out.equals(out2)

    def test_chunk_on_split(self, terra15_file_spool):
        """Ensure chunking which creates a slice at the end time works."""
        # this spool was selected because I first observed the issue in it.
        df = terra15_file_spool.get_contents()
        dur = (df["time_max"] - df["time_min"]).iloc[0]
        seg_len = dur / 3
        dt = df["time_step"].iloc[0]
        chunker = ChunkManager(keep_partial=True, time=seg_len)
        _, chunk_df = chunker.chunk(df)
        duration = chunk_df["time_max"] - chunk_df["time_min"]
        assert duration.sum() == ((seg_len - dt) * 3)
        assert len(duration) == 3
        assert (duration > np.timedelta64(0, "s")).all()

    def test_nan_in_df(self, contiguous_df):
        """Ensure contiguous df with nan inside still works."""
        df = contiguous_df.copy()
        # Adding null values on row 3
        df.loc[3, "time_min"] = dc.to_datetime64("NaT")
        # Which means new time should start in row 4 because of the gap.
        expected_start = df.loc[4, "time_min"]
        chunker = ChunkManager(keep_partial=True, time=dc.to_timedelta64(15))
        _, chunk_df = chunker.chunk(df)
        assert expected_start in set(chunk_df["time_min"])

    def test_all_nan(self, contiguous_df):
        """Ensure when all NaNs are encountered the chunked df is empty."""
        nat = dc.to_datetime64("NaT")
        df = contiguous_df.assign(time_min=nat, time_max=nat)
        chunker = ChunkManager(time=dc.to_timedelta64(1.2))
        _, chunk_df = chunker.chunk(df)
        assert chunk_df.empty

    def test_nan_in_sample_ok(self, contiguous_df):
        """Ensure a NaN in the sampling rate is ok."""
        df = contiguous_df.assign(time_step=dc.to_timedelta64("NaT"))
        dur = (df["time_max"] - df["time_min"]).iloc[0]
        chunker = ChunkManager(time=dc.to_timedelta64(dur / 2))
        _, chunk_df = chunker.chunk(df)
        assert isinstance(chunk_df, pd.DataFrame)
        assert len(chunk_df) == 2 * len(contiguous_df)
        assert np.all(pd.isnull(chunk_df["time_step"]))


class TestChunkExceptions:
    """Tests for various exceptions from the chunk manager."""

    def test_raises_overlap_no_chunksize(self):
        """Specifying an overlap and no chunk size should raise."""
        with pytest.raises(ParameterError, match="used for merging"):
            ChunkManager(time=None, overlap=10)

    def test_raises_zero_length_multiple_kwargs(self):
        """Ensure multiple kwargs raises nice error."""
        with pytest.raises(ParameterError, match="along one dimension"):
            ChunkManager(time=10, distance=1)

    def test_raises_zero_length_chunk(self):
        """Ensure zero length chunk raises."""
        with pytest.raises(ParameterError, match="must be greater than 0"):
            ChunkManager(time=0)

    def test_raises_invalid_key_in_kwargs(self, contiguous_df):
        """Ensure an invalid key in kwargs raises an error."""
        chunk_manager = ChunkManager(Time=10)
        chunk_manager.patch = type("Patch", (object,), {"dims": ["time", "distance"]})()
        with pytest.raises(ParameterError, match="Cannot chunk spool or"):
            chunk_manager.chunk(contiguous_df)


class TestChunkToMerge:
    """Tests for using chunking to merge contiguous, or overlapping, data."""

    @pytest.fixture()
    def gapy_df(self, contiguous_df):
        """Create a dataframe with gaps."""
        df = contiguous_df.copy()
        df["time_max"] -= df["time_step"] * 15
        return df

    @pytest.fixture()
    def gapy_df_unordered(self, gapy_df):
        """Create a dataframe with gaps that is not sorted by starttime."""
        inds = np.random.RandomState(42).permutation(gapy_df.index)
        return gapy_df.loc[inds].reset_index(drop=True)

    def test_chunk_can_merge(self, contiguous_df):
        """Ensure chunk can be used to merge unspecified segment lengths."""
        cm = ChunkManager(time=None)
        _, out = cm.chunk(contiguous_df)
        assert len(out) == 1
        assert out["time_min"].min() == contiguous_df["time_min"].min()

    def test_doesnt_merge_gappy_df(self, gapy_df):
        """Ensure the gappy dataframe doesn't get merged."""
        cm = ChunkManager(time=None)
        _, out = cm.chunk(gapy_df)
        assert len(gapy_df) == len(out)
        expected_durations = gapy_df["time_max"] - gapy_df["time_min"]
        durations = out["time_max"] - out["time_min"]
        assert expected_durations.equals(durations)

    def test_doesnt_merge_unordered_gappy_df(self, gapy_df_unordered):
        """Ensure the gappy dataframe doesn't get merged."""
        df = gapy_df_unordered
        cm = ChunkManager(time=None)
        _, out = cm.chunk(df)
        assert len(df) == len(out)
        expected_durations = df["time_max"] - df["time_min"]
        durations = out["time_max"] - out["time_min"]
        assert expected_durations.equals(durations)

    def test_modified_flag_after_merge(self, contiguous_df):
        """Test that the modified flag shows False for simple merge."""
        cm = ChunkManager(time=None)
        # Need to remove overlapping sample so these really are contiguous
        # with no overlaps.
        contiguous_df = contiguous_df.assign(
            time_max=lambda x: x["time_max"] - x["time_step"]
        )
        source, current = cm.chunk(contiguous_df)
        inst_df = cm.get_instruction_df(source, current)
        assert len(current) == 1
        assert current["time_min"].min() == contiguous_df["time_min"].min()
        assert not inst_df["_modified"].any()


class TestInstructionDF:
    """Sanity checks on intermediary df."""

    def test_indices(self, contiguous_df):
        """Ensure the input/output index belong to input/output df."""
        chunker = ChunkManager(overlap=0, time=10)
        in_df, out_df = chunker.chunk(contiguous_df)
        instruction = chunker.get_instruction_df(in_df, out_df)
        # ensure the source index is set as the index of the instruction_df
        assert instruction.index.name == "source_index"
        assert set(instruction.index).issubset(set(contiguous_df.index))
        assert set(instruction["current_index"]).issubset(set(out_df.index))

    def test_different_group_columns(self, contiguous_df_two_stations):
        """Ensure instruction df honors differences in group columns."""
        df = contiguous_df_two_stations
        chunker = ChunkManager(
            overlap=0,
            time=10,
            group_columns=("station",),
            keep_partial=True,
        )
        in_df, out_df = chunker.chunk(df)
        instruction = chunker.get_instruction_df(in_df, out_df)
        # ensure each output has exactly one station.
        for _current_index, sub in instruction.groupby("current_index"):
            source = df.loc[sub.index]
            # there should only be on station in the source for this group
            unique_stations = source["station"].unique()
            assert len(unique_stations) == 1
        # ensure all stations are present.
        used = in_df.loc[instruction.index]
        assert set(used["station"]) == set(in_df["station"])
        assert set(used["_group"]) == set(in_df["_group"])

    def test_modified_flag_if_chunked(self, contiguous_df):
        """Ensure the modified flag shows up for modified rows."""
        df = contiguous_df
        chunker = ChunkManager(
            overlap=0,
            time=5,
            group_columns=("station",),
            keep_partial=True,
        )
        in_df, out_df = chunker.chunk(df)
        instruction = chunker.get_instruction_df(in_df, out_df)
        assert instruction["_modified"].all()

    def test_modified_flag_no_chunk(self, contiguous_df):
        """Ensure the rows that don't change limits aren't modified."""
        time_diff = contiguous_df["time_max"] - contiguous_df["time_min"]
        df = contiguous_df.assign(time_max=lambda x: (x["time_max"] - x["time_step"]))
        chunker = ChunkManager(
            overlap=0,
            time=time_diff.iloc[0],
            group_columns=("station",),
            keep_partial=True,
        )
        in_df, out_df = chunker.chunk(df)

        assert (out_df[sorted(out_df.columns)]).equals(in_df[sorted(in_df.columns)])
        instruction = chunker.get_instruction_df(in_df, out_df)
        assert not instruction["_modified"].any()



"""Tests for example fetching."""

from __future__ import annotations

import numpy as np
import pytest

import dascore as dc
from dascore.examples import EXAMPLE_PATCHES
from dascore.exceptions import UnknownExampleError
from dascore.utils.time import to_float


class TestGetExamplePatch:
    """Test suite for `get_example_patch`."""

    def test_default(self):
        """Ensure calling get_example_patch with no args returns patch."""
        patch = dc.get_example_patch()
        assert isinstance(patch, dc.Patch)

    def test_raises_on_bad_key(self):
        """Ensure a bad key raises expected error."""
        with pytest.raises(UnknownExampleError, match="No example patch"):
            dc.get_example_patch("NotAnExampleRight????")

    def test_data_file_name(self):
        """Ensure get_example_spool works on a datafile."""
        spool = dc.get_example_spool("dispersion_event.h5")
        assert isinstance(spool, dc.BaseSpool)

    @pytest.mark.parametrize("name", EXAMPLE_PATCHES)
    def test_load_example_patch(self, name):
        """Ensure the registered example patches can all be loaded."""
        patch = dc.get_example_patch(name)
        assert isinstance(patch, dc.Patch)


class TestGetExampleSpool:
    """Test suite for `get_example_spool`."""

    def test_default(self):
        """Ensure calling get_example_spool with no args returns a Spool."""
        patch = dc.get_example_spool()
        assert isinstance(patch, dc.BaseSpool)

    def test_raises_on_bad_key(self):
        """Ensure a bad key raises expected error."""
        with pytest.raises(UnknownExampleError, match="No example spool"):
            dc.get_example_spool("NotAnExampleRight????")

    def test_data_file_name(self):
        """Ensure get_example_spool works on a datafile."""
        spool = dc.get_example_spool("dispersion_event.h5")
        assert isinstance(spool, dc.BaseSpool)


class TestRickerMoveout:
    """Tests for Ricker moveout patch."""

    def test_moveout(self):
        """Ensure peaks of ricker wavelet line up with expected moveout."""
        velocity = 100
        patch = dc.get_example_patch("ricker_moveout", velocity=velocity)
        argmaxes = np.argmax(patch.data, axis=0)
        peak_times = patch.get_coord("time").values[argmaxes]
        moveout = to_float(peak_times - np.min(peak_times))
        distances = patch.get_coord("distance").values
        expected_moveout = distances / velocity
        assert np.allclose(moveout, expected_moveout)


class TestDeltaPatch:
    """Tests for the delta_patch example."""

    @pytest.mark.parametrize("invalid_dim", ["inv_dim", "", None, 123, 1.1])
    def test_delta_patch_invalid_dim(self, invalid_dim):
        """
        Test that passing an invalid dimension value raises a ValueError.
        """
        msg = "with 'time' and 'distance'"
        with pytest.raises(ValueError, match=msg):
            dc.get_example_patch("delta_patch", dim=invalid_dim)

    @pytest.mark.parametrize("dim", ["time", "distance"])
    def test_delta_patch_structure(self, dim):
        """Test that the delta_patch returns a Patch with correct structure."""
        patch = dc.get_example_patch("delta_patch", dim=dim)
        assert isinstance(patch, dc.Patch), "delta_patch should return a Patch instance"

        dims = patch.dims
        assert (
            "time" in dims and "distance" in dims
        ), "Patch must have 'time' and 'distance' dimensions"

    @pytest.mark.parametrize("dim", ["time", "distance"])
    def test_delta_patch_delta_location(self, dim):
        """
        Ensures the delta is at the center of the chosen dimension and zeros elsewhere.
        """
        # The default shape from the function signature: shape=(10, 200)
        # If dim="time", we end up with a single (distance=0) trace => shape (200,)
        # If dim="distance", we end up with a single (time=0) trace => shape (10,)
        patch = dc.get_example_patch("delta_patch", dim=dim)
        data = patch.squeeze().data

        # The expected midpoint and verify single delta at center
        mid_idx = len(data) // 2

        assert data[mid_idx] == 1.0, "Expected a unit delta at the center"
        # Check all other samples are zero
        # Replace the center value with zero and ensure all zeros remain
        test_data = np.copy(data)
        test_data[mid_idx] = 0
        assert np.allclose(
            test_data, 0
        ), "All other samples should be zero except the center"

    @pytest.mark.parametrize("dim", ["time", "distance"])
    def test_delta_patch_with_patch(self, dim):
        """Test passing an existing patch to delta_patch and ensure delta is applied."""
        # Create a base patch
        base_patch = dc.get_example_patch("random_das", shape=(5, 50))
        # Apply the delta_patch function with the existing patch
        delta_applied_patch = dc.get_example_patch(
            "delta_patch", dim=dim, patch=base_patch
        )

        assert isinstance(delta_applied_patch, dc.Patch), "Should return a Patch"
        data = delta_applied_patch.squeeze().data

        # Check that only the center value is one and others are zero
        mid_idx = len(data) // 2
        assert data[mid_idx] == 1.0, "Center sample should be 1.0"
        test_data = np.copy(data)
        test_data[mid_idx] = 0
        assert np.allclose(
            test_data, 0
        ), "All other samples should be zero except the center"

    @pytest.mark.parametrize("dim", ["lag_time", "distance"])
    def test_delta_patch_with_3d_patch(self, dim):
        """Test passing a 3D patch."""
        # Create a base patch
        base_patch = dc.get_example_patch("sin_wav")
        base_patch_3d = base_patch.correlate(distance=[2], samples=True)
        # Apply the delta_patch function with the existing patch
        delta_applied_patch = dc.get_example_patch(
            "delta_patch", dim=dim, patch=base_patch_3d
        )
        assert isinstance(delta_applied_patch, dc.Patch), "Should return a Patch"
        data = delta_applied_patch.squeeze().data
        # Check that only the center value is one and others are zero
        mid_idx = len(data) // 2
        assert data[mid_idx] == 1.0, "Center sample should be 1.0"
        test_data = np.copy(data)
        test_data[mid_idx] = 0
        assert np.allclose(
            test_data, 0
        ), "All other samples should be zero except the center"
